# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common flags for the consumers subcommand group."""

import re

from googlecloudsdk.api_lib.services import services_util
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.calliope.concepts import deps
from googlecloudsdk.command_lib.util import completers
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs
from googlecloudsdk.core import properties

_SERVICES_LEGACY_LIST_COMMAND = ('services list --format=disable '
                                 '--flatten=serviceName')
_SERVICES_LIST_COMMAND = ('beta services list --format=disable '
                          '--flatten=config.name')

_OPERATION_NAME_RE = re.compile(r'operations/(?P<namespace>\w+)\.(?P<id>.*)')


class ConsumerServiceCompleter(completers.ListCommandCompleter):

  def __init__(self, **kwargs):
    super(ConsumerServiceCompleter, self).__init__(
        collection=services_util.SERVICES_COLLECTION,
        list_command=_SERVICES_LIST_COMMAND,
        flags=['enabled'],
        **kwargs)


class ConsumerServiceLegacyCompleter(completers.ListCommandCompleter):

  def __init__(self, **kwargs):
    super(ConsumerServiceLegacyCompleter, self).__init__(
        collection=services_util.SERVICES_COLLECTION,
        list_command=_SERVICES_LEGACY_LIST_COMMAND,
        flags=['enabled'],
        **kwargs)


def operation_flag(suffix='to act on'):
  return base.Argument(
      'operation', help='The name of the operation {0}.'.format(suffix))


def get_operation_namespace(op_name):
  match = _OPERATION_NAME_RE.match(op_name)
  if not match:
    raise arg_parsers.ArgumentTypeError("Invalid value '{0}': {1}".format(
        op_name, 'Operation format should be operations/namespace.id'))
  return match.group('namespace')


def consumer_service_flag(suffix='to act on', flag_name='service'):
  return base.Argument(
      flag_name,
      # This makes the argument required (at least one)
      nargs='+',
      completer=ConsumerServiceCompleter,
      help='The name of the service(s) {0}.'.format(suffix))


def single_consumer_service_flag(suffix='to act on', flag_name='service'):
  return base.Argument(
      flag_name,
      completer=ConsumerServiceCompleter,
      help='The name of the service {0}.'.format(suffix))


def available_service_flag(suffix='to act on', flag_name='service'):
  # NOTE: Because listing available services often forces the tab completion
  #       code to timeout, this flag will not enable tab completion.
  return base.Argument(
      flag_name,
      # This makes the argument required (at least one)
      nargs='+',
      help='The name of the service(s) {0}.'.format(suffix))


def service_flag(suffix='to act on', flag_name='service'):
  return base.Argument(
      flag_name, help='The name of the service {0}.'.format(suffix)
  )


def mcp_content_security_provider_flag(
    suffix='to act on', flag_name='mcp_content_security_provider'
):
  return base.Argument(
      flag_name,
      help='The MCP content security provider {0}.'.format(suffix),
  )


# TODO(b/321801975) Link public documentation about consumer policy format
# after public preview.
def consumer_policy_file_flag():
  return base.Argument(
      '--consumer-policy-file',
      required=True,
      help=(
          'Path to the file that contains the consumer policy in the YAML'
          ' format.'
      ),
  )


def bypass_dependency_check():
  return base.Argument(
      '--bypass-dependency-check',
      action='store_true',
      help=(
          ' This flag controls dependency management within the consumer'
          ' policy. If false, dependencies are enforced. When adding a new'
          ' service, it verifies that all the services dependencies'
          ' are already present in the policy. When removing a service, the'
          ' system will ensure that no other services in the policy depend on'
          ' it. If the policy lacks any required dependencies, the update will'
          ' not succeed, and an error will be returned.'
      ),
  )


def bypass_api_usage_check():
  return base.Argument(
      '--bypass-api-usage-check',
      action='store_true',
      help=(
          ' If true, the system will bypass usage checks for services'
          ' that are being removed. If false, the system will check if'
          ' the service to be removed was used in the last 30 days or enabled'
          ' in the last 3 days. If so, an error will be returned.'
      ),
  )


def skip_mcp_endpoint_check_flag(parser):
  return base.Argument(
      '--skip-mcp-endpoint-check',
      action='store_true',
      help=(
          ' If true, the system will bypass the check for MCP endpoint while'
          ' enabling a service.'
      ),
      hidden=True,
  ).AddToParser(parser)


def _create_key_resource_arg(help_txt, api_version, required=True):
  return presentation_specs.ResourcePresentationSpec(
      'key', _get_key_resource_spec(api_version), help_txt, required=required
  )


def _get_key_resource_spec(api_version):
  """Return the resource specification for a key."""
  if api_version == 'v2':
    return concepts.ResourceSpec(
        'apikeys.projects.locations.keys',
        api_version=api_version,
        resource_name='key',
        disable_auto_completers=True,
        keysId=_key_attribute_config(),
        locationsId=_location_attribute_config(),
        projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG)
  else:
    return concepts.ResourceSpec(
        'apikeys.projects.keys',
        api_version=api_version,
        resource_name='key',
        disable_auto_completers=True,
        keysId=_key_attribute_config(),
        projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG)


def _key_attribute_config():
  return concepts.ResourceParameterAttributeConfig(
      name='key', help_text='Id of the key')


def _location_attribute_config():
  return concepts.ResourceParameterAttributeConfig(
      name='location',
      help_text='Location of the key.',
      fallthroughs=[
          deps.Fallthrough(
              function=lambda: 'global',
              hint='location will default to {}'.format('global'),
              active=True,
              plural=False)
      ])


def key_flag(parser, suffix='to act on', api_version='v2', required=True):
  return concept_parsers.ConceptParser(
      [
          _create_key_resource_arg(
              help_txt='The name of the key {0}.'.format(suffix),
              api_version=api_version,
              required=required,
          )
      ]
  ).AddToParser(parser)


def display_name_flag(parser, suffix='to act on'):
  base.Argument(
      '--display-name',
      help='Display name of the key {0}.'.format(suffix)).AddToParser(parser)


def key_id_flag(parser, suffix='to act on'):
  base.Argument(
      '--key-id', help='User-specified ID of the key {0}.'.format(suffix)
  ).AddToParser(parser)


def service_account_flag(parser):
  base.Argument(
      '--service-account',
      help=(
          'The email of the service account the key is bound to. If this field'
          ' is specified, the key is a service account bound key and auth'
          ' enabled.'
      ),
  ).AddToParser(parser)


def add_key_undelete_args(parser):
  """Adds args for api-keys undelete command."""
  undelete_set_group = parser.add_mutually_exclusive_group(required=True)
  key_flag(
      undelete_set_group, suffix='to undelete', api_version='v2', required=False
  )
  _key_string_flag(undelete_set_group)


def skip_dependency_flag(parser):
  base.Argument(
      '--skip-dependency',
      action='store_true',
      help=(
          'If set, the dependencies of the service to be enabled will not be'
          ' enabled.'
      ),
  ).AddToParser(parser)


def validate_only_args(parser, suffix='to act on'):
  base.Argument(
      '--validate-only',
      action='store_true',
      help=(
          "Validate the {} action, but don't actually perform it".format(suffix)
      ),
  ).AddToParser(parser)


def _bypass_dependency_service_check_flag(parser):
  base.Argument(
      '--bypass-dependency-service-check',
      action='store_true',
      help=(
          'If specified, the disable call will bypass the check for'
          ' dependencies and the dependencies will remain enabled.'
      ),
  ).AddToParser(parser)


def bypass_api_usage_check_flag(parser):
  base.Argument(
      '--bypass-api-usage-check',
      action='store_true',
      help=(
          'If specified, the system will bypass usage checks for services that'
          ' are being removed. Otherwise, the system will check if the service'
          ' to be removed was used in the last 30 days or enabled in the last 3'
          ' days. If so, the system will return an error.'
      ),
  ).AddToParser(parser)


def _disable_dependency_services_flag(parser):
  base.Argument(
      '--disable-dependency-services',
      action='store_true',
      help=(
          ' If specified, the disable call will proceed disabling the service'
          ' and all the enabled services depend on the service to be disabled'
      ),
  ).AddToParser(parser)


def add_dependency_check_args(parser):
  """Adds resource args for command."""
  dependent_check_group = parser.add_mutually_exclusive_group(required=False)
  _disable_dependency_services_flag(dependent_check_group)
  _bypass_dependency_service_check_flag(dependent_check_group)


def add_resource_args(parser):
  """Adds resource args for command."""
  resource_group = parser.add_mutually_exclusive_group(required=False)
  _project_id_flag(resource_group)
  _folder_id_flag(resource_group)
  _orgnaization_id_flag(resource_group)


def add_key_update_args(parser):
  """Adds args for api-keys update command."""
  update_set_restriction_group = parser.add_mutually_exclusive_group(
      required=False
  )
  _add_clear_restrictions_arg(update_set_restriction_group)
  restriction_group = update_set_restriction_group.add_argument_group()
  client_restriction_group = restriction_group.add_mutually_exclusive_group()
  _allowed_referrers_arg(client_restriction_group)
  _allowed_ips_arg(client_restriction_group)
  _allowed_bundle_ids(client_restriction_group)
  _allowed_application(client_restriction_group)
  _api_targets_arg(restriction_group)
  update_set_annotation_group = parser.add_mutually_exclusive_group(
      required=False
  )
  _annotations(update_set_annotation_group)
  _add_clear_annotations_arg(update_set_annotation_group)


def add_key_create_args(parser):
  """Add args for api-keys create command."""
  restriction_group = parser.add_argument_group()
  client_restriction_group = restriction_group.add_mutually_exclusive_group()
  _allowed_referrers_arg(client_restriction_group)
  _allowed_ips_arg(client_restriction_group)
  _allowed_bundle_ids(client_restriction_group)
  _allowed_application(client_restriction_group)
  _api_targets_arg(restriction_group)
  _annotations(parser)


def _add_clear_restrictions_arg(parser):
  base.Argument(
      '--clear-restrictions',
      action='store_true',
      help='If set, clear all restrictions on the key.').AddToParser(parser)


def _add_clear_annotations_arg(parser):
  base.Argument(
      '--clear-annotations',
      action='store_true',
      help='If set, clear all annotations on the key.').AddToParser(parser)


def _allowed_referrers_arg(parser):
  base.Argument(
      '--allowed-referrers',
      default=[],
      type=arg_parsers.ArgList(),
      metavar='ALLOWED_REFERRERS',
      help='A list of regular expressions for the referrer URLs that are '
      'allowed to make API calls with this key.').AddToParser(parser)


def _allowed_ips_arg(parser):
  base.Argument(
      '--allowed-ips',
      default=[],
      type=arg_parsers.ArgList(),
      metavar='ALLOWED_IPS',
      help='A list of the caller IP addresses that are allowed to make API '
      'calls with this key.').AddToParser(parser)


def _allowed_bundle_ids(parser):
  base.Argument(
      '--allowed-bundle-ids',
      default=[],
      metavar='ALLOWED_BUNDLE_IDS',
      type=arg_parsers.ArgList(),
      help='iOS app\'s bundle ids that are allowed to use the key.'
  ).AddToParser(parser)


def _allowed_application(parser):
  base.Argument(
      '--allowed-application',
      type=arg_parsers.ArgDict(
          spec={
              'sha1_fingerprint': str,
              'package_name': str
          },
          required_keys=['sha1_fingerprint', 'package_name'],
          max_length=2),
      metavar='sha1_fingerprint=SHA1_FINGERPRINT,package_name=PACKAGE_NAME',
      action='append',
      help=('Repeatable. Specify multiple allowed applications. '
            'The accepted keys are `sha1_fingerprint` and `package_name`.'
           )).AddToParser(parser)


def _annotations(parser):
  base.Argument(
      '--annotations',
      metavar='KEY=VALUE',
      type=arg_parsers.ArgDict(),
      required=False,
      help=("""Annotations are key resource. Specify annotations as
            a key-value dictionary for small amounts of arbitrary client data.
            """)).AddToParser(parser)


def _api_targets_arg(parser):
  base.Argument(
      '--api-target',
      type=arg_parsers.ArgDict(
          spec={
              'service': str,
              'methods': arg_parsers.ArgList(custom_delim_char=':'),
          },
          required_keys=['service'],
          min_length=1,
      ),
      metavar='service=SERVICE[,methods=METHOD:...]',
      action='append',
      help="""\
      Repeatable. Specify service and optionally one or multiple specific
      methods. Both fields are case insensitive.
      Methods can be specified as a colon-separated list (e.g.,
      `--api-target="service=foo.service.com,methods=method1:method2"`),
      or using `--flags-file` (see $ gcloud topic flags-file for details).
      """,
  ).AddToParser(parser)


def _key_string_flag(parser):
  base.Argument('--key-string', help='Key String of the key.').AddToParser(
      parser
  )


def _project_id_flag(parser):
  base.Argument(
      '--project',
      metavar='PROJECT_ID',
      help="""\
The Google Cloud project ID to use for this invocation. If
omitted, then the current project is assumed; the current project can
be listed using `gcloud config list --format='text(core.project)'`
and can be set using `gcloud config set project PROJECTID`.

`--project` and its fallback `{core_project}` property play two roles
in the invocation. It specifies the project of the resource to
operate on. It also specifies the project for API enablement check,
quota, and billing. To specify a different project for quota and
billing, use `--billing-project` or `{billing_project}` property.
    """.format(
          core_project=properties.VALUES.core.project,
          billing_project=properties.VALUES.billing.quota_project,
      ),
  ).AddToParser(parser)


def _folder_id_flag(parser):
  base.Argument(
      '--folder',
      metavar='FOLDER_ID',
      help='The Google Cloud Platform folder ID to use for this invocation.',
  ).AddToParser(parser)


def _orgnaization_id_flag(parser):
  base.Argument(
      '--organization',
      metavar='ORGANIZATION_ID',
      help=(
          'The Google Cloud Platform organization ID to use for this'
          ' invocation.'
      ),
  ).AddToParser(parser)



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convenience functions for dealing with instance templates."""


from googlecloudsdk.api_lib.compute import alias_ip_range_utils
from googlecloudsdk.api_lib.compute import constants
from googlecloudsdk.api_lib.compute import image_utils
from googlecloudsdk.api_lib.compute import instance_utils
from googlecloudsdk.api_lib.compute import kms_utils
from googlecloudsdk.api_lib.compute import utils
from googlecloudsdk.api_lib.compute.instances.create import utils as create_utils
from googlecloudsdk.command_lib.compute import scope as compute_scope
from googlecloudsdk.command_lib.compute.networks.subnets import flags as subnet_flags
from googlecloudsdk.command_lib.util.args import labels_util
from googlecloudsdk.core import properties

EPHEMERAL_ADDRESS = object()


def CreateNetworkInterfaceMessage(
    resources,
    scope_lister,
    messages,
    network,
    private_ip,
    subnet_region,
    subnet,
    address,
    alias_ip_ranges_string=None,
    alias_ipv6_ranges_string=None,
    network_tier=None,
    stack_type=None,
    ipv6_network_tier=None,
    nic_type=None,
    ipv6_public_ptr_domain=None,
    ipv6_address=None,
    ipv6_prefix_length=None,
    external_ipv6_address=None,
    external_ipv6_prefix_length=None,
    internal_ipv6_address=None,
    internal_ipv6_prefix_length=None,
    network_attachment=None,
    network_queue_count=None,
    vlan=None,
    igmp_query=None,
    enable_vpc_scoped_dns=None,
    service_class_id=None,
    support_alias_ipv6_ranges=False,
    dns64_eligible=None,
    nat64_eligible=None,
):
  """Creates and returns a new NetworkInterface message.

  Args:
    resources: generates resource references,
    scope_lister: function, provides scopes for prompting subnet region,
    messages: GCE API messages,
    network: network,
    private_ip: IPv4 internal IP address to assign to the instance.
    subnet_region: region for subnetwork,
    subnet: regional subnetwork,
    address: specify static address for instance template * None - no address, *
      EPHEMERAL_ADDRESS - ephemeral address, * string - address name to be
      fetched from GCE API.
    alias_ip_ranges_string: command line string specifying a list of alias IP
      ranges.
    alias_ipv6_ranges_string: command line string specifying a list of alias IP
      v6 ranges.
    network_tier: specify network tier for instance template * None - no network
      tier * PREMIUM - network tier being PREMIUM * SELECT - network tier being
      SELECT * STANDARD - network tier being STANDARD
    stack_type: identify whether IPv6 features are enabled * IPV4_ONLY - can
      only have IPv4 address * IPV4_IPV6 - can have both IPv4 and IPv6 address *
      IPV6_ONLY - can only have IPv6 address
    ipv6_network_tier: specify network tier for IPv6 access config * PREMIUM -
      network tier being PREMIUM * STANDARD - network tier being STANDARD
    nic_type: specify the type of NetworkInterface Controller * GVNIC *
      VIRTIO_NET
    ipv6_public_ptr_domain: a string represents the custom PTR domain assigned
      to the interface.
    ipv6_address: a string represents the external IPv6 address reserved to the
      interface.
    ipv6_prefix_length: a string represents the external IPv6 address prefix
      length reserved to the interface.
    external_ipv6_address: a string represents the external IPv6 address
      reserved to the interface.
    external_ipv6_prefix_length: a string represents the external IPv6 address
      prefix length reserved to the interface.
    internal_ipv6_address: a string represents the internal IPv6 address
      reserved to the interface.
    internal_ipv6_prefix_length:  the internal IPv6 address prefix length of the
      internal IPv6 address reserved to the interface.
    network_attachment: URL of a network attachment to connect the interface to.
    network_queue_count: the number of queues assigned to the network interface.
    vlan: the VLAN tag of the network interface.
    igmp_query: the IGMP query mode of the network interface.
    enable_vpc_scoped_dns: If True, indicates that this network interface can be
      used to send DNS queries to the VPC network's scope.
    service_class_id: Producer Service's Service class Id for the region of this
      network interface. Can only be used with network_attachment.
    support_alias_ipv6_ranges: Indicates whether setting alias IPv6 ranges on
      network interfaces is supported.
    dns64_eligible: If True, indicates that this network interface is eligible
      for DNS64.
    nat64_eligible: If True, indicates that this network interface is eligible
      for NAT64.

  Returns:
    network_interface: a NetworkInterface message object
  """
  # By default interface is attached to default network. If network or subnet
  # are specified they're used instead.
  network_interface = messages.NetworkInterface()
  if subnet is not None:
    subnet_ref = subnet_flags.SubnetworkResolver().ResolveResources(
        [subnet],
        compute_scope.ScopeEnum.REGION,
        subnet_region,
        resources,
        scope_lister=scope_lister)[0]
    network_interface.subnetwork = subnet_ref.SelfLink()
  if network is not None:
    network_ref = resources.Parse(
        network,
        params={'project': properties.VALUES.core.project.GetOrFail},
        collection='compute.networks')
    network_interface.network = network_ref.SelfLink()
  # A network interface referencing a network attachment cannot have
  # a subnetwork or a network set.
  elif subnet is None and network_attachment is None:
    network_ref = resources.Parse(
        constants.DEFAULT_NETWORK,
        params={'project': properties.VALUES.core.project.GetOrFail},
        collection='compute.networks')
    network_interface.network = network_ref.SelfLink()

  if private_ip is not None:
    network_interface.networkIP = private_ip

  if stack_type is not None:
    network_interface.stackType = (
        messages.NetworkInterface.StackTypeValueValuesEnum(stack_type))

  if address and stack_type != 'IPV6_ONLY':
    access_config = messages.AccessConfig(
        name=constants.DEFAULT_ACCESS_CONFIG_NAME,
        type=messages.AccessConfig.TypeValueValuesEnum.ONE_TO_ONE_NAT)

    # If the user provided an external IP, populate the access
    # config with it.
    if address != EPHEMERAL_ADDRESS:
      access_config.natIP = address

    if network_tier is not None:
      access_config.networkTier = (
          messages.AccessConfig.NetworkTierValueValuesEnum(network_tier))

    network_interface.accessConfigs = [access_config]

  ipv6_access_config = None
  if ipv6_network_tier is not None or ipv6_public_ptr_domain is not None:
    ipv6_access_config = messages.AccessConfig(
        name=constants.DEFAULT_IPV6_ACCESS_CONFIG_NAME,
        type=messages.AccessConfig.TypeValueValuesEnum.DIRECT_IPV6)
    network_interface.ipv6AccessConfigs = [ipv6_access_config]

  if ipv6_network_tier is not None:
    ipv6_access_config.networkTier = (
        messages.AccessConfig.NetworkTierValueValuesEnum(ipv6_network_tier))

  if ipv6_public_ptr_domain is not None:
    ipv6_access_config.publicPtrDomainName = ipv6_public_ptr_domain

  # external_ipv6_address has higher priority than ipv6_address.
  if external_ipv6_address is None:
    external_ipv6_address = ipv6_address

  # external_ipv6_prefix_length has higher priority than ipv6_prefix_length.
  if external_ipv6_prefix_length is None:
    external_ipv6_prefix_length = ipv6_prefix_length

  if external_ipv6_address is not None:
    if ipv6_access_config is None:
      ipv6_access_config = messages.AccessConfig(
          name=constants.DEFAULT_IPV6_ACCESS_CONFIG_NAME,
          type=messages.AccessConfig.TypeValueValuesEnum.DIRECT_IPV6)
      network_interface.ipv6AccessConfigs = [ipv6_access_config]

    ipv6_access_config.externalIpv6 = external_ipv6_address

  if external_ipv6_prefix_length is not None:
    if ipv6_access_config is None:
      ipv6_access_config = messages.AccessConfig(
          name=constants.DEFAULT_IPV6_ACCESS_CONFIG_NAME,
          type=messages.AccessConfig.TypeValueValuesEnum.DIRECT_IPV6)
      network_interface.ipv6AccessConfigs = [ipv6_access_config]

    ipv6_access_config.externalIpv6PrefixLength = external_ipv6_prefix_length

  if internal_ipv6_address is not None:
    network_interface.ipv6Address = internal_ipv6_address

  if internal_ipv6_prefix_length is not None:
    network_interface.internalIpv6PrefixLength = internal_ipv6_prefix_length

  if support_alias_ipv6_ranges:
    if alias_ip_ranges_string:
      network_interface.aliasIpRanges = (
          alias_ip_range_utils.CreateAliasIpRangeMessagesFromString(
              messages, False, alias_ip_ranges_string
          )
      )
    if alias_ipv6_ranges_string:
      network_interface.aliasIpv6Ranges = (
          alias_ip_range_utils.CreateAliasIpv6RangeMessagesFromString(
              messages, False, alias_ipv6_ranges_string
          )
      )
  else:
    if alias_ip_ranges_string:
      network_interface.aliasIpRanges = (
          alias_ip_range_utils.CreateAliasIpRangeMessagesFromStringOld(
              messages, False, alias_ip_ranges_string
          )
      )

  if nic_type is not None:
    network_interface.nicType = (
        messages.NetworkInterface.NicTypeValueValuesEnum(nic_type))

  if network_attachment is not None:
    network_interface.networkAttachment = network_attachment

  if network_queue_count is not None:
    network_interface.queueCount = network_queue_count

  if vlan is not None:
    network_interface.vlan = vlan

  if igmp_query is not None:
    network_interface.igmpQuery = (
        messages.NetworkInterface.IgmpQueryValueValuesEnum(igmp_query)
    )

  if enable_vpc_scoped_dns:
    network_interface.enableVpcScopedDns = enable_vpc_scoped_dns

  if service_class_id:
    network_interface.serviceClassId = service_class_id

  if dns64_eligible:
    network_interface.dns64Eligible = dns64_eligible
  if nat64_eligible:
    network_interface.nat64Eligible = nat64_eligible

  return network_interface


def CreateNetworkInterfaceMessages(
    resources,
    scope_lister,
    messages,
    network_interface_arg,
    subnet_region,
    support_enable_vpc_scoped_dns=False,
    support_alias_ipv6_ranges=False,
    support_dns64_eligible=False,
    support_nat64_eligible=False,
):
  """Create network interface messages.

  Args:
    resources: generates resource references,
    scope_lister: function, provides scopes for prompting subnet region,
    messages: creates resources.
    network_interface_arg: CLI argument specifying network interfaces.
    subnet_region: region of the subnetwork.
    support_enable_vpc_scoped_dns: Indicates whether setting enable vpc scoped
      dns on network interfaces is supported.
    support_alias_ipv6_ranges: Indicates whether setting alias IPv6 ranges on
      network interfaces is supported.
    support_dns64_eligible: If True, indicates that this network interface is
      eligible for DNS64.
    support_nat64_eligible: If True, indicates that this network interface is
      eligible for NAT64.

  Returns:
    list, items are NetworkInterfaceMessages.
  """
  result = []
  if network_interface_arg:
    for interface in network_interface_arg:
      address = interface.get('address', None)
      # A network interface referencing a network attachment cannot have
      # an access config.
      has_no_address = 'no-address' in interface or interface.get(
          'network-attachment', None)
      # pylint: disable=g-explicit-bool-comparison
      if address == '' or (address is None and (not has_no_address)):
        address = EPHEMERAL_ADDRESS

      network_tier = interface.get('network-tier', None)
      nic_type = interface.get('nic-type', None)
      enable_vpc_scoped_dns = None
      if support_enable_vpc_scoped_dns:
        enable_vpc_scoped_dns = 'enable-vpc-scoped-dns' in interface
      service_class_id = interface.get('service-class-id', None)
      dns64_eligible = (
          'dns64-eligible' in interface if support_dns64_eligible else None
      )
      nat64_eligible = (
          'nat64-eligible' in interface if support_nat64_eligible else None
      )

      result.append(
          CreateNetworkInterfaceMessage(
              resources,
              scope_lister,
              messages,
              interface.get('network', None),
              interface.get('private-network-ip', None),
              subnet_region,
              interface.get('subnet', None),
              address,
              interface.get('aliases', None),
              interface.get('ipv6-aliases', None),
              network_tier,
              nic_type=nic_type,
              stack_type=interface.get('stack-type', None),
              ipv6_network_tier=interface.get('ipv6-network-tier', None),
              ipv6_public_ptr_domain=interface.get(
                  'ipv6-public-ptr-domain', None
              ),
              ipv6_address=interface.get('ipv6-address', None),
              ipv6_prefix_length=interface.get('ipv6-prefix-length', None),
              external_ipv6_address=interface.get(
                  'external-ipv6-address', None
              ),
              external_ipv6_prefix_length=interface.get(
                  'external-ipv6-prefix-length', None
              ),
              internal_ipv6_address=interface.get(
                  'internal-ipv6-address', None
              ),
              internal_ipv6_prefix_length=interface.get(
                  'internal-ipv6-prefix-length', None
              ),
              network_attachment=interface.get('network-attachment'),
              network_queue_count=interface.get('queue-count', None),
              vlan=interface.get('vlan', None),
              igmp_query=interface.get('igmp-query', None),
              enable_vpc_scoped_dns=enable_vpc_scoped_dns,
              service_class_id=service_class_id,
              support_alias_ipv6_ranges=support_alias_ipv6_ranges,
              dns64_eligible=dns64_eligible,
              nat64_eligible=nat64_eligible,
          )
      )
  return result


def CreateDiskMessages(
    args,
    client,
    resources,
    project,
    image_uri,
    boot_disk_size_gb=None,
    create_boot_disk=False,
    support_kms=False,
    support_multi_writer=False,
    match_container_mount_disks=False,
    support_replica_zones=True,
    support_disk_labels=False,
):
  """Create disk messages for a single instance template."""
  container_mount_disk = (
      args.container_mount_disk if match_container_mount_disks else [])

  persistent_disks = (
      CreatePersistentAttachedDiskMessages(
          client.messages,
          args.disk or [],
          container_mount_disk=container_mount_disk))

  persistent_create_disks = CreatePersistentCreateDiskMessages(
      client,
      resources,
      project,
      getattr(args, 'create_disk', []),
      support_kms=support_kms,
      support_multi_writer=support_multi_writer,
      support_replica_zones=support_replica_zones,
      support_disk_labels=support_disk_labels,
  )

  if create_boot_disk:
    boot_disk_list = [
        CreateDefaultBootAttachedDiskMessage(
            messages=client.messages,
            disk_type=args.boot_disk_type,
            disk_device_name=args.boot_disk_device_name,
            disk_auto_delete=args.boot_disk_auto_delete,
            disk_size_gb=boot_disk_size_gb,
            image_uri=image_uri,
            kms_args=args,
            support_kms=support_kms,
            disk_provisioned_iops=args.boot_disk_provisioned_iops,
            disk_provisioned_throughput=args.boot_disk_provisioned_throughput,
            disk_interface=args.boot_disk_interface,
        )
    ]
  elif persistent_create_disks and persistent_create_disks[0].boot:
    boot_disk_list = [persistent_create_disks.pop(0)]
  elif persistent_disks and persistent_disks[0].boot:
    boot_disk_list = [persistent_disks.pop(0)]
  else:
    boot_disk_list = []

  local_nvdimms = create_utils.CreateLocalNvdimmMessages(
      args,
      resources,
      client.messages,
  )

  local_ssds = create_utils.CreateLocalSsdMessages(
      args,
      resources,
      client.messages,
  )

  return (
      boot_disk_list
      + persistent_disks
      + persistent_create_disks
      + local_nvdimms
      + local_ssds
  )


def CreatePersistentAttachedDiskMessages(
    messages, disks, container_mount_disk=None
):
  """Returns a list of AttachedDisk messages and the boot disk's reference.

  Args:
    messages: GCE API messages,
    disks: disk objects - contains following properties
             * name - the name of disk,
             * mode - 'rw' (R/W), 'ro' (R/O) access mode,
             * boot - whether it is a boot disk,
             * auto-delete - whether disks is deleted when VM is deleted,
             * device-name - device name on VM.
    container_mount_disk: list of disks to be mounted to container, if any.

  Returns:
    list of API messages for attached disks
  """

  disks_messages = []
  for disk in disks:
    name = disk['name']
    # Resolves the mode.
    mode_value = disk.get('mode', 'rw')
    if mode_value == 'rw':
      mode = messages.AttachedDisk.ModeValueValuesEnum.READ_WRITE
    else:
      mode = messages.AttachedDisk.ModeValueValuesEnum.READ_ONLY

    boot = disk.get('boot', False)
    auto_delete = disk.get('auto-delete', False)
    device_name = instance_utils.GetDiskDeviceName(disk, name,
                                                   container_mount_disk)

    attached_disk = messages.AttachedDisk(
        autoDelete=auto_delete,
        boot=boot,
        deviceName=device_name,
        mode=mode,
        source=name,
        type=messages.AttachedDisk.TypeValueValuesEnum.PERSISTENT)

    if disk.get('interface'):
      if disk.get('interface') == 'SCSI':
        interface = messages.AttachedDisk.InterfaceValueValuesEnum.SCSI
      else:
        interface = messages.AttachedDisk.InterfaceValueValuesEnum.NVME
      attached_disk.interface = interface

    # The boot disk must end up at index 0.
    if boot:
      disks_messages = [attached_disk] + disks_messages
    else:
      disks_messages.append(attached_disk)

  return disks_messages


def CreatePersistentCreateDiskMessages(
    client,
    resources,
    user_project,
    create_disks,
    support_kms=False,
    container_mount_disk=None,
    support_multi_writer=False,
    support_replica_zones=True,
    support_disk_labels=False,
):
  """Returns a list of AttachedDisk messages.

  Args:
    client: Compute client adapter
    resources: Compute resources registry
    user_project: name of user project
    create_disks: disk objects - contains following properties * name - the name
      of disk, * description - an optional description for the disk, * mode -
      'rw' (R/W), 'ro' (R/O) access mode, * size - the size of the disk, *
      provisioned-iops - Indicates how many IOPS must be provisioned for the
      disk. * provisioned-throughput - Indicates how much throughput is
      provisioned for the disks. * type - the type of the disk (HDD or SSD), *
      image - the name of the image to initialize from, * image-family - the
      image family name, * image-project - the project name that has the image,
      * auto-delete - whether disks is deleted when VM is deleted ('yes' if
      True), * device-name - device name on VM, * disk-resource-policy -
      resource policies applied to disk. * storage-pool: the storage pool in
      which the new disk is created.
    support_kms: if KMS is supported
    container_mount_disk: list of disks to be mounted to container, if any.
    support_multi_writer: if multi writer disks are supported.
    support_replica_zones: True if we allow creation of regional disks.
    support_disk_labels: True if we allow adding disk labels.

  Returns:
    list of API messages for attached disks
  """

  disks_messages = []
  messages = client.messages
  for disk in create_disks or []:
    name = disk.get('name')
    # Resolves the mode.
    mode_value = disk.get('mode', 'rw')
    if mode_value == 'rw':
      mode = client.messages.AttachedDisk.ModeValueValuesEnum.READ_WRITE
    else:
      mode = client.messages.AttachedDisk.ModeValueValuesEnum.READ_ONLY

    auto_delete = disk.get('auto-delete', False)
    boot = disk.get('boot', False)
    disk_size_gb = utils.BytesToGb(disk.get('size'))
    img = disk.get('image')
    img_family = disk.get('image-family')
    img_project = disk.get('image-project')

    image_uri = None
    if img or img_family:
      image_expander = image_utils.ImageExpander(client, resources)
      image_uri, _ = image_expander.ExpandImageFlag(
          user_project=user_project,
          image=img,
          image_family=img_family,
          image_project=img_project,
          return_image_resource=False)

    disk_key = None
    if support_kms:
      disk_key = kms_utils.MaybeGetKmsKeyFromDict(disk, client.messages,
                                                  disk_key)

    device_name = instance_utils.GetDiskDeviceName(disk, name,
                                                   container_mount_disk)

    init_params = client.messages.AttachedDiskInitializeParams(
        diskName=name,
        description=disk.get('description'),
        sourceImage=image_uri,
        diskSizeGb=disk_size_gb,
        diskType=disk.get('type'),
        provisionedIops=disk.get('provisioned-iops'))

    replica_zones = disk.get('replica-zones')
    if support_replica_zones and replica_zones:
      normalized_zones = []
      for zone in replica_zones:
        zone_ref = resources.Parse(
            zone, collection='compute.zones', params={'project': user_project})
        normalized_zones.append(zone_ref.SelfLink())
      init_params.replicaZones = normalized_zones

    policies = disk.get('disk-resource-policy')
    if policies:
      init_params.resourcePolicies = policies

    multi_writer = disk.get('multi-writer')
    if support_multi_writer and multi_writer:
      init_params.multiWriter = True

    disk_architecture = disk.get('architecture')
    if disk_architecture:
      init_params.architecture = (
          messages.AttachedDiskInitializeParams.ArchitectureValueValuesEnum(
              disk_architecture
          )
      )

    if support_disk_labels:
      dict_labels = labels_util.ValidateAndParseLabels(disk.get('labels'))
      if dict_labels:
        labels_value = messages.AttachedDiskInitializeParams.LabelsValue()
        labels_value.additionalProperties = [
            labels_value.AdditionalProperty(key=key, value=value)
            for key, value in dict_labels.items()
        ]

        init_params.labels = labels_value

    init_params.provisionedThroughput = disk.get('provisioned-throughput')

    init_params.storagePool = disk.get('storage-pool')

    create_disk = client.messages.AttachedDisk(
        autoDelete=auto_delete,
        boot=boot,
        deviceName=device_name,
        initializeParams=init_params,
        mode=mode,
        type=client.messages.AttachedDisk.TypeValueValuesEnum.PERSISTENT,
        diskEncryptionKey=disk_key,
    )
    if disk.get('interface'):
      if disk.get('interface') == 'SCSI':
        interface = messages.AttachedDisk.InterfaceValueValuesEnum.SCSI
      else:
        interface = messages.AttachedDisk.InterfaceValueValuesEnum.NVME
      create_disk.interface = interface

    # The boot disk must end up at index 0.
    if boot:
      disks_messages = [create_disk] + disks_messages
    else:
      disks_messages.append(create_disk)

  return disks_messages


def CreateDefaultBootAttachedDiskMessage(
    messages,
    disk_type,
    disk_device_name,
    disk_auto_delete,
    disk_size_gb,
    image_uri,
    kms_args=None,
    support_kms=False,
    disk_provisioned_iops=None,
    disk_provisioned_throughput=None,
    disk_interface=None,
):
  """Returns an AttachedDisk message for creating a new boot disk."""
  disk_key = None

  if support_kms:
    disk_key = kms_utils.MaybeGetKmsKey(
        kms_args, messages, disk_key, boot_disk_prefix=True)

  initialize_params = messages.AttachedDiskInitializeParams(
      sourceImage=image_uri, diskSizeGb=disk_size_gb, diskType=disk_type)

  if disk_provisioned_iops is not None:
    initialize_params.provisionedIops = disk_provisioned_iops
  if disk_provisioned_throughput is not None:
    initialize_params.provisionedThroughput = disk_provisioned_throughput

  boot_attached_disk = messages.AttachedDisk(
      autoDelete=disk_auto_delete,
      boot=True,
      deviceName=disk_device_name,
      initializeParams=initialize_params,
      mode=messages.AttachedDisk.ModeValueValuesEnum.READ_WRITE,
      type=messages.AttachedDisk.TypeValueValuesEnum.PERSISTENT,
      diskEncryptionKey=disk_key,
  )
  if disk_interface:
    if disk_interface == 'SCSI':
      interface = messages.AttachedDisk.InterfaceValueValuesEnum.SCSI
    else:
      interface = messages.AttachedDisk.InterfaceValueValuesEnum.NVME
    boot_attached_disk.interface = interface
  return boot_attached_disk


def CreateAcceleratorConfigMessages(messages, accelerator):
  """Returns a list of accelerator config messages for Instance Templates.

  Args:
    messages: tracked GCE API messages.
    accelerator: accelerator object with the following properties:
        * type: the accelerator's type.
        * count: the number of accelerators to attach. Optional, defaults to 1.

  Returns:
    a list of accelerator config messages that specify the type and number of
    accelerators to attach to an instance.
  """
  if accelerator is None:
    return []

  accelerator_type = accelerator['type']
  accelerator_count = int(accelerator.get('count', 1))
  accelerator_config = messages.AcceleratorConfig(
      acceleratorType=accelerator_type, acceleratorCount=accelerator_count)
  return [accelerator_config]



# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reducer functions to generate instance props from prior state and flags."""


import argparse
import datetime

from googlecloudsdk.api_lib.sql import api_util as common_api_util
from googlecloudsdk.api_lib.sql import constants
from googlecloudsdk.api_lib.sql import exceptions as sql_exceptions
from googlecloudsdk.api_lib.sql import instances as api_util
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.core import properties
from googlecloudsdk.core.util import files
import six


def ActiveDirectoryConfig(
    sql_messages,
    domain=None,
    mode=None,
    dns_servers=None,
    admin_credential_secret_key=None,
    organizational_unit=None,
):
  """Generates the Active Directory configuration for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    domain: string, the Active Directory domain value.
    mode: string, the Active Directory mode value.
    dns_servers: list of strings, the list of dns servers.
    admin_credential_secret_key: string, the name of the admin credential secret
      manager key.
    organizational_unit: string, the organizational unit value.

  Returns:
    sql_messages.SqlActiveDirectoryConfig object.
  """

  should_generate_config = any([
      domain is not None,
      mode is not None,
      dns_servers is not None,
      admin_credential_secret_key is not None,
      organizational_unit is not None,
  ])
  if not should_generate_config:
    return None

  config = sql_messages.SqlActiveDirectoryConfig()
  if domain is not None:
    config.domain = domain
  if admin_credential_secret_key is not None:
    config.adminCredentialSecretName = admin_credential_secret_key
  if mode is not None:
    config.mode = sql_messages.SqlActiveDirectoryConfig.ModeValueValuesEnum.lookup_by_name(
        mode.upper()
    )
  if dns_servers:
    config.dnsServers = dns_servers
  if organizational_unit is not None:
    config.organizationalUnit = organizational_unit
  return config


def SqlServerAuditConfig(sql_messages,
                         bucket=None,
                         retention_interval=None,
                         upload_interval=None):
  """Generates the Audit configuration for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    bucket: string, the GCS bucket name.
    retention_interval: duration, how long to keep generated audit files.
    upload_interval: duration, how often to upload generated audit files.

  Returns:
    sql_messages.SqlServerAuditConfig object.
  """

  if bucket is None and retention_interval is None and upload_interval is None:
    return None

  config = sql_messages.SqlServerAuditConfig()
  if bucket is not None:
    config.bucket = bucket
  if retention_interval is not None:
    config.retentionInterval = six.text_type(retention_interval) + 's'
  if upload_interval is not None:
    config.uploadInterval = six.text_type(upload_interval) + 's'

  return config


def SqlServerEntraIdConfig(
    sql_messages,
    tenant_id=None,
    application_id=None,
    clear_entraid_config=False,
):
  """Generates the Entra ID configuration for the SQL Server instance.

  Args:
    sql_messages: module, The messages module that should be used.
    tenant_id: string, the Entra ID tenant ID.
    application_id: string, the Entra ID application ID.
    clear_entraid_config: If true, clear the Entra ID config.

  Returns:
    sql_messages.SqlServerEntraIdConfig object.
  """

  if clear_entraid_config:
    return sql_messages.SqlServerEntraIdConfig(
        tenantId='',
        applicationId='',
    )

  if (tenant_id is None and application_id is None):
    return None

  config = sql_messages.SqlServerEntraIdConfig()
  if tenant_id is not None:
    config.tenantId = tenant_id
  if application_id is not None:
    config.applicationId = application_id
  return config


def BackupConfiguration(
    sql_messages,
    instance=None,
    backup_enabled=None,
    backup_location=None,
    backup_start_time=None,
    enable_bin_log=None,
    enable_point_in_time_recovery=None,
    retained_backups_count=None,
    retained_transaction_log_days=None,
    patch_request=False,
):
  """Generates the backup configuration for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    instance: sql_messages.DatabaseInstance, the original instance, if the
      previous state is needed.
    backup_enabled: boolean, True if backup should be enabled.
    backup_location: string, location where to store backups by default.
    backup_start_time: string, start time of backup specified in 24-hour format.
    enable_bin_log: boolean, True if binary logging should be enabled.
    enable_point_in_time_recovery: boolean, True if point-in-time recovery
      (using write-ahead log archiving) should be enabled.
    retained_backups_count: int, how many backups to keep stored.
    retained_transaction_log_days: int, how many days of transaction logs to
      keep stored.
    patch_request: boolean, True if this is a patch request.

  Returns:
    sql_messages.BackupConfiguration object, or None

  Raises:
    ToolException: Bad combination of arguments.
  """
  should_generate_config = any([
      backup_location is not None,
      backup_start_time,
      enable_bin_log is not None,
      enable_point_in_time_recovery is not None,
      retained_backups_count is not None,
      retained_transaction_log_days is not None,
      not backup_enabled,
  ])

  if not should_generate_config:
    return None

  if not instance or not instance.settings.backupConfiguration:
    backup_config = sql_messages.BackupConfiguration(
        kind='sql#backupConfiguration',
        startTime='00:00',
        enabled=backup_enabled)
  else:
    backup_config = instance.settings.backupConfiguration

  gcbdr_managed = (
      backup_config.backupTier
      == sql_messages.BackupConfiguration.BackupTierValueValuesEnum.ENHANCED
  )

  if patch_request and gcbdr_managed:
    raise sql_exceptions.ArgumentError(
        'Backup configuration cannot be changed for instances with a BackupDR'
        ' backup plan attached.'
    )

  if backup_location is not None:
    backup_config.location = backup_location
    backup_config.enabled = True
  if backup_start_time:
    backup_config.startTime = backup_start_time
    backup_config.enabled = True

  if retained_backups_count is not None:
    backup_retention_settings = (
        backup_config.backupRetentionSettings or
        sql_messages.BackupRetentionSettings())
    backup_retention_settings.retentionUnit = sql_messages.BackupRetentionSettings.RetentionUnitValueValuesEnum.COUNT
    backup_retention_settings.retainedBackups = retained_backups_count

    backup_config.backupRetentionSettings = backup_retention_settings
    backup_config.enabled = True

  if retained_transaction_log_days is not None:
    backup_config.transactionLogRetentionDays = retained_transaction_log_days
    backup_config.enabled = True

  if not backup_enabled:
    if (backup_location is not None or backup_start_time or
        retained_backups_count is not None or
        retained_transaction_log_days is not None):
      raise sql_exceptions.ArgumentError(
          'Argument --no-backup not allowed with --backup-location, '
          '--backup-start-time, --retained-backups-count, or '
          '--retained-transaction-log-days')
    backup_config.enabled = False

  if enable_bin_log is not None:
    backup_config.binaryLogEnabled = enable_bin_log

  if enable_point_in_time_recovery is not None:
    backup_config.pointInTimeRecoveryEnabled = enable_point_in_time_recovery
    if backup_config.replicationLogArchivingEnabled is not None:
      backup_config.replicationLogArchivingEnabled = (
          enable_point_in_time_recovery
      )

  # retainedTransactionLogDays is only valid when we have transaction logs,
  # i.e, have binlog or pitr.
  if (
      retained_transaction_log_days
      and not backup_config.binaryLogEnabled
      and not backup_config.pointInTimeRecoveryEnabled
  ):
    raise sql_exceptions.ArgumentError(
        'Argument --retained-transaction-log-days only valid when '
        'transaction logs are enabled. To enable transaction logs, use '
        '--enable-bin-log for MySQL, and use --enable-point-in-time-recovery '
        'for Postgres and SQL Server.'
    )

  return backup_config


def DatabaseFlags(sql_messages,
                  settings=None,
                  database_flags=None,
                  clear_database_flags=False):
  """Generates the database flags for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    settings: sql_messages.Settings, the original settings, if the previous
      state is needed.
    database_flags: dict of flags.
    clear_database_flags: boolean, True if flags should be cleared.

  Returns:
    list of sql_messages.DatabaseFlags objects
  """
  updated_flags = []
  if database_flags:
    for (name, value) in sorted(database_flags.items()):
      updated_flags.append(sql_messages.DatabaseFlags(name=name, value=value))
  elif clear_database_flags:
    updated_flags = []
  elif settings:
    updated_flags = settings.databaseFlags

  return updated_flags


def Tags(sql_messages, tags=None):
  """Generates the tags for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    tags: list of tags.

  Returns:
    list of sql_messages.Tags objects
  """
  updated_tags = []
  if tags:
    for tag in tags:
      updated_tags.append(sql_messages.Tags(tag=tag))

  return updated_tags


def MaintenanceWindow(sql_messages,
                      instance,
                      maintenance_release_channel=None,
                      maintenance_window_day=None,
                      maintenance_window_hour=None):
  """Generates the maintenance window for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    instance: sql_messages.DatabaseInstance, The original instance, if it might
      be needed to generate the maintenance window.
    maintenance_release_channel: string, which channel's updates to apply.
    maintenance_window_day: string, maintenance window day of week.
    maintenance_window_hour: int, maintenance window hour of day.

  Returns:
    sql_messages.MaintenanceWindow or None

  Raises:
    argparse.ArgumentError: no maintenance window specified.
  """
  channel = maintenance_release_channel
  day = maintenance_window_day
  hour = maintenance_window_hour
  if not any([channel, day, hour]):
    return None

  maintenance_window = sql_messages.MaintenanceWindow(
      kind='sql#maintenanceWindow')

  # If there's no existing maintenance window,
  # both or neither of day and hour must be set.
  if (not instance or not instance.settings or
      not instance.settings.maintenanceWindow):
    if ((day is None and hour is not None) or
        (hour is None and day is not None)):
      raise argparse.ArgumentError(
          None, 'There is currently no maintenance window on the instance. '
          'To add one, specify values for both day, and hour.')

  if channel:
    # Map UI name to API name.
    names = {
        'week5':
            sql_messages.MaintenanceWindow.UpdateTrackValueValuesEnum.week5,
        'production':
            sql_messages.MaintenanceWindow.UpdateTrackValueValuesEnum.stable,
        'preview':
            sql_messages.MaintenanceWindow.UpdateTrackValueValuesEnum.canary
    }
    maintenance_window.updateTrack = names[channel]
  if day:
    # Map day name to number.
    day_num = arg_parsers.DayOfWeek.DAYS.index(day)
    if day_num == 0:
      day_num = 7
    maintenance_window.day = day_num
  if hour is not None:  # must execute on hour = 0
    maintenance_window.hour = hour
  return maintenance_window


def DenyMaintenancePeriod(sql_messages,
                          instance,
                          deny_maintenance_period_start_date=None,
                          deny_maintenance_period_end_date=None,
                          deny_maintenance_period_time='00:00:00'):
  """Generates the deny maintenance period for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    instance: sql_messages.DatabaseInstance, The original instance, if it might
      be needed to generate the deny maintenance period.
    deny_maintenance_period_start_date: date, Date when the deny maintenance
      period begins, i.e., 2020-11-01.
    deny_maintenance_period_end_date: date, Date when the deny maintenance
      period ends, i.e., 2021-01-10.
    deny_maintenance_period_time: Time when the deny maintenance period
      starts/ends, i.e., 05:00:00.

  Returns:
    sql_messages.DenyMaintenancePeriod or None

  Raises:
    argparse.ArgumentError: invalid deny maintenance period specified.
  """
  old_deny_maintenance_period = None
  if (instance and instance.settings and
      instance.settings.denyMaintenancePeriods and
      instance.settings.denyMaintenancePeriods[0]):
    old_deny_maintenance_period = instance.settings.denyMaintenancePeriods[0]

  deny_maintenance_period = sql_messages.DenyMaintenancePeriod()

  if old_deny_maintenance_period:
    # if there is deny maintenance period specified for the instance
    deny_maintenance_period = old_deny_maintenance_period
    if deny_maintenance_period_start_date:
      ValidateDate(deny_maintenance_period_start_date)
      deny_maintenance_period.startDate = deny_maintenance_period_start_date

    if deny_maintenance_period_end_date:
      ValidateDate(deny_maintenance_period_end_date)
      deny_maintenance_period.endDate = deny_maintenance_period_end_date

    if deny_maintenance_period_time:
      ValidTime(deny_maintenance_period_time)
      deny_maintenance_period.time = deny_maintenance_period_time
  else:
    if not (deny_maintenance_period_start_date and
            deny_maintenance_period_end_date):
      raise argparse.ArgumentError(
          None, 'There is no deny maintenance period on the instance.'
          ' To add one, specify values for both start date and end date.')
    ValidateDate(deny_maintenance_period_start_date)
    deny_maintenance_period.startDate = deny_maintenance_period_start_date

    ValidateDate(deny_maintenance_period_end_date)
    deny_maintenance_period.endDate = deny_maintenance_period_end_date

    if deny_maintenance_period_time:
      ValidTime(deny_maintenance_period_time)
      deny_maintenance_period.time = deny_maintenance_period_time

  return deny_maintenance_period


def ValidTime(s):
  try:
    datetime.datetime.strptime(s, '%H:%M:%S')
  except ValueError:
    raise argparse.ArgumentError(
        None, 'Invalid time value. The format should be HH:mm:SS.')


def ValidateDate(s):
  try:
    return datetime.datetime.strptime(s, '%Y-%m-%d')
  except ValueError:
    try:
      return datetime.datetime.strptime(s, '%m-%d')
    except ValueError:
      raise argparse.ArgumentError(
          None, 'Invalid date value. The format should be yyyy-mm-dd or mm-dd.')


def InsightsConfig(sql_messages,
                   insights_config_query_insights_enabled=None,
                   insights_config_enhanced_query_insights_enabled=None,
                   insights_config_query_string_length=None,
                   insights_config_record_application_tags=None,
                   insights_config_record_client_address=None,
                   insights_config_query_plans_per_minute=None):
  """Generates the insights config for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    insights_config_query_insights_enabled: boolean, True if query insights
      should be enabled.
    insights_config_enhanced_query_insights_enabled: boolean, True if enhanced
      query insights should be enabled.
    insights_config_query_string_length: number, length of the query string to
      be stored.
    insights_config_record_application_tags: boolean, True if application tags
      should be recorded.
    insights_config_record_client_address: boolean, True if client address
      should be recorded.
    insights_config_query_plans_per_minute: number, number of query plans to
      sample every minute.

  Returns:
    sql_messages.InsightsConfig or None
  """

  should_generate_config = any([
      insights_config_query_insights_enabled is not None,
      insights_config_enhanced_query_insights_enabled is not None,
      insights_config_query_string_length is not None,
      insights_config_record_application_tags is not None,
      insights_config_record_client_address is not None,
      insights_config_query_plans_per_minute is not None,
  ])
  if not should_generate_config:
    return None

  # Config exists, generate insights config.
  insights_config = sql_messages.InsightsConfig()
  if insights_config_query_insights_enabled is not None:
    insights_config.queryInsightsEnabled = (
        insights_config_query_insights_enabled)
  if insights_config_enhanced_query_insights_enabled is not None:
    insights_config.enhancedQueryInsightsEnabled = (
        insights_config_enhanced_query_insights_enabled)
  if insights_config_query_string_length is not None:
    insights_config.queryStringLength = insights_config_query_string_length
  if insights_config_record_application_tags is not None:
    insights_config.recordApplicationTags = (
        insights_config_record_application_tags)
  if insights_config_record_client_address is not None:
    insights_config.recordClientAddress = insights_config_record_client_address
  if insights_config_query_plans_per_minute is not None:
    insights_config.queryPlansPerMinute = insights_config_query_plans_per_minute

  return insights_config


def DbAlignedAtomicWritesConfig(sql_messages, db_aligned_atomic_writes=None):
  """Generates the db aligned atomic writes Config for the instance."""
  if db_aligned_atomic_writes is None:
    return None
  return sql_messages.DbAlignedAtomicWritesConfig(
      dbAlignedAtomicWrites=db_aligned_atomic_writes
  )


def ConnectionPoolConfig(
    sql_messages,
    enable_connection_pooling=None,
    connection_pool_flags=None,
    clear_connection_pool_flags=None,
    current_config=None,
):
  """Generates the connection pooling config for the instance."""

  # Skip generate new config if no config field is requested to be updated.
  if all([
      enable_connection_pooling is None,
      connection_pool_flags is None,
      clear_connection_pool_flags is None,
  ]):
    return None

  connection_pool_config = current_config or sql_messages.ConnectionPoolConfig()
  if enable_connection_pooling is not None:
    connection_pool_config.connectionPoolingEnabled = enable_connection_pooling

  if connection_pool_flags is not None:
    updated_flags = []
    for name, value in sorted(connection_pool_flags.items()):
      updated_flags.append(
          sql_messages.ConnectionPoolFlags(name=name, value=value)
      )
    connection_pool_config.flags = updated_flags
  elif clear_connection_pool_flags:
    connection_pool_config.flags = []

  return connection_pool_config


def ReadPoolAutoScaleConfig(
    sql_messages,
    auto_scale_enabled=None,
    auto_scale_min_node_count=None,
    auto_scale_max_node_count=None,
    auto_scale_target_metrics=None,
    auto_scale_disable_scale_in=None,
    auto_scale_in_cooldown_seconds=None,
    auto_scale_out_cooldown_seconds=None,
    current_config=None,
):
  """Generates the read pool auto-scale config for the instance."""
  if all([
      auto_scale_enabled is None,
      auto_scale_min_node_count is None,
      auto_scale_max_node_count is None,
      auto_scale_target_metrics is None,
      auto_scale_disable_scale_in is None,
      auto_scale_in_cooldown_seconds is None,
      auto_scale_out_cooldown_seconds is None,
  ]):
    return None
  read_pool_auto_scale_config = (
      current_config or sql_messages.ReadPoolAutoScaleConfig()
  )
  if auto_scale_enabled is not None:
    if auto_scale_enabled:
      read_pool_auto_scale_config.enabled = True
    else:
      # If auto-scale is disabled, clear the config.
      return sql_messages.ReadPoolAutoScaleConfig(enabled=False)
  if auto_scale_min_node_count is not None:
    read_pool_auto_scale_config.minNodeCount = auto_scale_min_node_count
  if auto_scale_max_node_count is not None:
    read_pool_auto_scale_config.maxNodeCount = auto_scale_max_node_count
  if auto_scale_target_metrics is not None:
    read_pool_auto_scale_config.targetMetrics = []
    for (
        metric,
        value,
    ) in auto_scale_target_metrics.items():
      read_pool_auto_scale_config.targetMetrics.append(
          sql_messages.TargetMetric(
              metric=metric,
              targetValue=value,
          )
      )
  if auto_scale_disable_scale_in is not None:
    read_pool_auto_scale_config.disableScaleIn = auto_scale_disable_scale_in
  if auto_scale_in_cooldown_seconds is not None:
    read_pool_auto_scale_config.scaleInCooldownSeconds = (
        auto_scale_in_cooldown_seconds
    )
  if auto_scale_out_cooldown_seconds is not None:
    read_pool_auto_scale_config.scaleOutCooldownSeconds = (
        auto_scale_out_cooldown_seconds
    )
  return read_pool_auto_scale_config


def _CustomMachineTypeString(cpu, memory_mib):
  """Creates a custom machine type from the CPU and memory specs.

  Args:
    cpu: the number of cpu desired for the custom machine type
    memory_mib: the amount of ram desired in MiB for the custom machine type
      instance

  Returns:
    The custom machine type name for the 'instance create' call
  """
  machine_type = 'db-custom-{0}-{1}'.format(cpu, memory_mib)
  return machine_type


def MachineType(instance=None, tier=None, memory=None, cpu=None):
  """Generates the machine type for the instance.

  Adapted from compute.

  Args:
    instance: sql_messages.DatabaseInstance, The original instance, if it might
      be needed to generate the machine type.
    tier: string, the v1 or v2 tier.
    memory: string, the amount of memory.
    cpu: int, the number of CPUs.

  Returns:
    A string representing the URL naming a machine-type.

  Raises:
    exceptions.RequiredArgumentException when only one of the two custom
        machine type flags are used, or when none of the flags are used.
    exceptions.InvalidArgumentException when both the tier and
        custom machine type flags are used to generate a new instance.
  """

  # Setting the machine type.
  machine_type = None
  if tier:
    machine_type = tier

  # Setting the specs for the custom machine.
  if cpu or memory:
    if not cpu:
      raise exceptions.RequiredArgumentException(
          '--cpu', 'Both [--cpu] and [--memory] must be '
          'set to create a custom machine type instance.')
    if not memory:
      raise exceptions.RequiredArgumentException(
          '--memory', 'Both [--cpu] and [--memory] must '
          'be set to create a custom machine type instance.')
    if tier:
      raise exceptions.InvalidArgumentException(
          '--tier', 'Cannot set both [--tier] and '
          '[--cpu]/[--memory] for the same instance.')
    custom_type_string = _CustomMachineTypeString(
        cpu,
        # Converting from B to MiB.
        memory // (2**20))

    # Updating the machine type that is set for the URIs.
    machine_type = custom_type_string

  # Reverting to default if creating instance and no flags are set.
  if not machine_type and not instance:
    machine_type = constants.DEFAULT_MACHINE_TYPE

  return machine_type


def OnPremisesConfiguration(sql_messages, source_ip_address, source_port):
  """Generates the external primary configuration for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    source_ip_address: string, the IP address of the external data source.
    source_port: number, the port number of the external data source.

  Returns:
    sql_messages.OnPremisesConfiguration object.
  """
  return sql_messages.OnPremisesConfiguration(
      kind='sql#onPremisesConfiguration',
      hostPort='{0}:{1}'.format(source_ip_address, source_port))


def PrivateNetworkUrl(network):
  """Generates the self-link of the instance's private network.

  Args:
    network: The ID of the network.

  Returns:
    string, the URL of the network.
  """
  client = common_api_util.SqlClient(common_api_util.API_VERSION_DEFAULT)
  network_ref = client.resource_parser.Parse(
      network,
      params={
          'project': properties.VALUES.core.project.GetOrFail,
      },
      collection='compute.networks')
  return network_ref.SelfLink()


def ReplicaConfiguration(sql_messages,
                         primary_username,
                         primary_password,
                         primary_dump_file_path,
                         primary_ca_certificate_path=None,
                         client_certificate_path=None,
                         client_key_path=None):
  """Generates the config for an external primary replica.

  Args:
    sql_messages: module, The messages module that should be used.
    primary_username: The username for connecting to the external instance.
    primary_password: The password for connecting to the external instance.
    primary_dump_file_path: ObjectReference, a wrapper for the URI of the Cloud
      Storage path containing the dumpfile to seed the replica with.
    primary_ca_certificate_path: The path to the CA certificate PEM file.
    client_certificate_path: The path to the client certificate PEM file.
    client_key_path: The path to the client private key PEM file.

  Returns:
    sql_messages.MySqlReplicaConfiguration object.
  """
  mysql_replica_configuration = sql_messages.MySqlReplicaConfiguration(
      kind='sql#mysqlReplicaConfiguration',
      username=primary_username,
      password=primary_password,
      dumpFilePath=primary_dump_file_path.ToUrl())
  if primary_ca_certificate_path:
    mysql_replica_configuration.caCertificate = files.ReadFileContents(
        primary_ca_certificate_path)
  if client_certificate_path:
    mysql_replica_configuration.clientCertificate = files.ReadFileContents(
        client_certificate_path)
  if client_key_path:
    mysql_replica_configuration.clientKey = files.ReadFileContents(
        client_key_path)
  return sql_messages.ReplicaConfiguration(
      kind='sql#demoteMasterMysqlReplicaConfiguration',
      mysqlReplicaConfiguration=mysql_replica_configuration)


def Region(specified_region, gce_zone, secondary_zone=None):
  """Generates the region string for the instance.

  Args:
    specified_region: string, the GCE region to create the instance in.
    gce_zone: string, the GCE zone to create the instance in.
    secondary_zone: string, the GCE zone to create the standby instance in.

  Returns:
    string, the region to create the instance in.
  """
  if gce_zone and secondary_zone:
    region_from_zone = api_util.GetRegionFromZone(gce_zone)
    region_from_secondary_zone = api_util.GetRegionFromZone(secondary_zone)
    if region_from_zone != region_from_secondary_zone:
      raise exceptions.ConflictingArgumentsException(
          'Zones in arguments --zone and --secondary-zone '
          'belong to different regions.')
  if gce_zone:
    derived_region = api_util.GetRegionFromZone(gce_zone)
    return derived_region
  return specified_region


def _ParseComplexity(sql_messages, complexity):
  if complexity:
    return sql_messages.PasswordValidationPolicy.ComplexityValueValuesEnum.lookup_by_name(
        complexity.upper())
  return None


def PasswordPolicy(
    sql_messages,
    password_policy_min_length=None,
    password_policy_complexity=None,
    password_policy_reuse_interval=None,
    password_policy_disallow_username_substring=None,
    password_policy_password_change_interval=None,
    enable_password_policy=None,
    clear_password_policy=None,
):
  """Generates or clears password policy for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    password_policy_min_length: int, Minimum number of characters allowed.
    password_policy_complexity: string, The complexity of the password.
    password_policy_reuse_interval: int, Number of previous passwords that
      cannot be reused.
    password_policy_disallow_username_substring: boolean, True if disallow
      username as a part of the password.
    password_policy_password_change_interval: duration, Minimum interval at
      which password can be changed.
    enable_password_policy: boolean, True if password validation policy is
      enabled.
    clear_password_policy: boolean, True if clear existing password policy.

  Returns:
    sql_messages.PasswordValidationPolicy or None
  """
  should_generate_policy = any([
      password_policy_min_length is not None,
      password_policy_complexity is not None,
      password_policy_reuse_interval is not None,
      password_policy_disallow_username_substring is not None,
      password_policy_password_change_interval is not None,
      enable_password_policy is not None,
  ])
  if not should_generate_policy or clear_password_policy:
    return None

  # Config exists, generate password policy.
  password_policy = sql_messages.PasswordValidationPolicy()

  if password_policy_min_length is not None:
    password_policy.minLength = password_policy_min_length
  if password_policy_complexity is not None:
    password_policy.complexity = _ParseComplexity(sql_messages,
                                                  password_policy_complexity)
  if password_policy_reuse_interval is not None:
    password_policy.reuseInterval = password_policy_reuse_interval
  if password_policy_disallow_username_substring is not None:
    password_policy.disallowUsernameSubstring = password_policy_disallow_username_substring
  if password_policy_password_change_interval is not None:
    password_policy.passwordChangeInterval = str(
        password_policy_password_change_interval) + 's'
  if enable_password_policy is not None:
    password_policy.enablePasswordPolicy = enable_password_policy

  return password_policy


def PscAutoConnections(
    sql_messages,
    psc_auto_connections=None,
):
  """Generates PSC auto connections for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    psc_auto_connections: dict of the allowed consumer projects and networks.

  Returns:
    list of sql_messages.PscAutoConnectionConfig objects

  Raises:
    exceptions.InvalidArgumentException when there is no valid network or
    project specified.
  """
  updated_psc_auto_connections = []
  for connection in psc_auto_connections:
    current_psc_auto_connection = sql_messages.PscAutoConnectionConfig()
    current_psc_auto_connection.consumerNetwork = connection.get('network')

    if project := connection.get('project'):
      current_psc_auto_connection.consumerProject = project
    else:
      client = common_api_util.SqlClient(common_api_util.API_VERSION_DEFAULT)
      network_ref = client.resource_parser.ParseRelativeName(
          current_psc_auto_connection.consumerNetwork,
          collection='compute.networks',
      )
      current_psc_auto_connection.consumerProject = network_ref.project

    if (
        current_psc_auto_connection.consumerProject
        and current_psc_auto_connection.consumerNetwork
    ):
      updated_psc_auto_connections.append(current_psc_auto_connection)
    else:
      raise exceptions.InvalidArgumentException(
          '--psc-auto-connections', 'PSC auto connection must have network '
          'specified.'
      )
  return updated_psc_auto_connections


def UncMappings(sql_messages, unc_mappings=None, clear_unc_mappings=False):
  """Generates the database flags for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    unc_mappings: input unc mappings from the user.
    clear_unc_mappings: boolean, True if unc mappings should be cleared.

  Returns:
    list of sql_messages.UncMapping objects
  """
  updated_mappings = []
  if unc_mappings:
    for mapping_str in unc_mappings:
      try:
        rest, mode_str = mapping_str.rsplit(':', 1)
        unc_path, gcs_path = rest.split('=', 1)
        updated_mappings.append(
            sql_messages.UncMapping(
                uncPath=unc_path,
                gcsPath=gcs_path,
                mode=sql_messages.UncMapping.ModeValueValuesEnum.lookup_by_name(
                    mode_str.upper()
                ),
            )
        )
      except ValueError as e:
        raise ValueError(
            "Invalid UNC mapping input. Expected 'unc-path=gcs-path:mode'."
        ) from e
      except KeyError as e:
        raise ValueError(
            "Invalid mode. Expected 'snapshot_read' or 'snapshot_write'."
        ) from e
  elif clear_unc_mappings:
    updated_mappings = []

  return updated_mappings


def PerformanceCaptureConfig(
    sql_messages,
    performance_capture_config=None,
    current_config=None,
):
  """Generates the performance capture config for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    performance_capture_config: dict of the performance capture config.
    current_config: sql_messages.PerformanceCaptureConfig, the original
      performance capture config, if the previous state is needed.

  Returns:
    sql_messages.PerformanceCaptureConfig object, or None

  Raises:
    exceptions.InvalidArgumentException: If an unknown key is provided in
      performance_capture_config.
  """

  if performance_capture_config is None:
    return None

  config = current_config or sql_messages.PerformanceCaptureConfig()

  # Map CLI keys to config attributes
  bool_keys = {
      'enabled': 'enabled',
  }
  int_keys = {
      'probing-interval-seconds': 'probingIntervalSeconds',
      'probe-threshold': 'probeThreshold',
      'running-threads-threshold': 'runningThreadsThreshold',
      'seconds-behind-source-threshold': 'secondsBehindSourceThreshold',
      'transaction-duration-threshold': 'transactionDurationThreshold',
      'cpu-utilization-threshold-percent': 'cpuUtilizationThresholdPercent',
      'memory-usage-threshold-percent': 'memoryUsageThresholdPercent',
      'transaction-lock-wait-threshold-count': (
          'transactionLockWaitThresholdCount'
      ),
      'semaphore-wait-threshold-count': 'semaphoreWaitThresholdCount',
      'history-list-length-threshold-count': 'historyListLengthThresholdCount',
      'transaction-kill-threshold-seconds': 'transactionKillThresholdSeconds',
  }
  list_keys = {
      'transaction-kill-excluded-user-hosts': (
          'transactionKillExcludedUserHosts'
      ),
  }
  enum_keys = {
      'transaction-kill-type': (
          'transactionKillType',
          sql_messages.PerformanceCaptureConfig.TransactionKillTypeValueValuesEnum,
      ),
  }

  for key, value in sorted(performance_capture_config.items()):
    if key in bool_keys:
      val_str = value.lower()
      if val_str not in ['true', 'false']:
        raise exceptions.InvalidArgumentException(
            '--performance-capture-config',
            f'Invalid boolean value for key [{key}]: {value}',
        )
      setattr(config, bool_keys[key], val_str == 'true')
    elif key in int_keys:
      try:
        setattr(config, int_keys[key], int(value))
      except ValueError as exc:
        raise exceptions.InvalidArgumentException(
            '--performance-capture-config',
            f'Invalid integer value for key [{key}]: {value}',
        ) from exc
    elif key in list_keys:
      if value is None:
        setattr(config, list_keys[key], [])
        continue
      if not isinstance(value, str):
        raise exceptions.InvalidArgumentException(
            '--performance-capture-config',
            f'Invalid value for key [{key}]: expected string.',
        )
      setattr(
          config,
          list_keys[key],
          [v.strip() for v in value.split(';')] if value else [],
      )
    elif key in enum_keys:
      attr_name, enum_cls = enum_keys[key]
      try:
        enum_val = enum_cls.lookup_by_name(value.upper())
        setattr(config, attr_name, enum_val)
      except KeyError as exc:
        raise exceptions.InvalidArgumentException(
            '--performance-capture-config',
            f'Invalid value for key [{key}]: {value}',
        ) from exc
    else:
      raise exceptions.InvalidArgumentException(
          '--performance-capture-config', f'Unknown key: {key}'
      )

  return config


def FinalBackupConfiguration(
    sql_messages,
    instance=None,
    final_backup_enabled=None,
    final_backup_retention_days=None,
):
  """Generates the Final Backup configuration for the instance.

  Args:
    sql_messages: module, The messages module that should be used.
    instance: sql_messages.DatabaseInstance, the original instance, if the
      previous state is needed.
    final_backup_enabled: boolean, True if final backup should be enabled.
    final_backup_retention_days: int, how many days to retain the final backup.

  Returns:
    sql_messages.FinalBackupConfiguration object, or None

  Raises:
    sql_exceptions.ArgumentError: Bad combination of arguments.
  """
  should_generate_config = any([
      final_backup_enabled is not None,
      final_backup_retention_days is not None,
  ])

  if not should_generate_config:
    return None

  # final_backup_enabled is explicitly set to False.
  # final_backup_retention_days should not be set using gcloud.
  if final_backup_enabled is not None and not final_backup_enabled:
    if final_backup_retention_days is not None:
      raise sql_exceptions.ArgumentError(
          'You cannot set final-backup-retention-days while final-backup field is disabled.'
      )

  if not instance or not instance.settings.finalBackupConfig:
    final_backup_config = sql_messages.FinalBackupConfig()
  else:
    final_backup_config = instance.settings.finalBackupConfig

  # Generate new final backup config based on the gcloud arguments.
  if final_backup_enabled is not None:
    final_backup_config.enabled = final_backup_enabled
  if final_backup_retention_days is not None:
    final_backup_config.retentionDays = final_backup_retention_days
    # Final backup enabled set to true if retention days specified.
    final_backup_config.enabled = True

  # final_backup_enabled is set to False, we need to cleanup the retention days.
  if final_backup_enabled is not None and not final_backup_enabled:
    final_backup_config.retentionDays = None

  return final_backup_config



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Resource definitions for Cloud Platform Apis generated from apitools."""

import enum


BASE_URL = 'https://apihub.googleapis.com/v1/'
DOCS_URL = 'https://cloud.google.com/apigee/docs/api-hub/what-is-api-hub'


class Collections(enum.Enum):
  """Collections for all supported apis."""

  PROJECTS = (
      'projects',
      'projects/{projectsId}',
      {},
      ['projectsId'],
      True
  )
  PROJECTS_LOCATIONS = (
      'projects.locations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_ADDONS = (
      'projects.locations.addons',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/addons/'
              '{addonsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_APIHUBINSTANCES = (
      'projects.locations.apiHubInstances',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/apiHubInstances/'
              '{apiHubInstancesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_APIS = (
      'projects.locations.apis',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/apis/{apisId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_APIS_VERSIONS = (
      'projects.locations.apis.versions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/apis/{apisId}/'
              'versions/{versionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_APIS_VERSIONS_DEFINITIONS = (
      'projects.locations.apis.versions.definitions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/apis/{apisId}/'
              'versions/{versionsId}/definitions/{definitionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_APIS_VERSIONS_OPERATIONS = (
      'projects.locations.apis.versions.operations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/apis/{apisId}/'
              'versions/{versionsId}/operations/{operationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_APIS_VERSIONS_SPECS = (
      'projects.locations.apis.versions.specs',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/apis/{apisId}/'
              'versions/{versionsId}/specs/{specsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_ATTRIBUTES = (
      'projects.locations.attributes',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/attributes/'
              '{attributesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_CURATIONS = (
      'projects.locations.curations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/curations/'
              '{curationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_DEPENDENCIES = (
      'projects.locations.dependencies',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/dependencies/'
              '{dependenciesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_DEPLOYMENTS = (
      'projects.locations.deployments',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/deployments/'
              '{deploymentsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_DISCOVEREDAPIOBSERVATIONS = (
      'projects.locations.discoveredApiObservations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'discoveredApiObservations/{discoveredApiObservationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_DISCOVEREDAPIOBSERVATIONS_DISCOVEREDAPIOPERATIONS = (
      'projects.locations.discoveredApiObservations.discoveredApiOperations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'discoveredApiObservations/{discoveredApiObservationsId}/'
              'discoveredApiOperations/{discoveredApiOperationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_EXTERNALAPIS = (
      'projects.locations.externalApis',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/externalApis/'
              '{externalApisId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_HOSTPROJECTREGISTRATIONS = (
      'projects.locations.hostProjectRegistrations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'hostProjectRegistrations/{hostProjectRegistrationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_OPERATIONS = (
      'projects.locations.operations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/operations/'
              '{operationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_PLUGINS = (
      'projects.locations.plugins',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/plugins/'
              '{pluginsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_PLUGINS_INSTANCES = (
      'projects.locations.plugins.instances',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/plugins/'
              '{pluginsId}/instances/{instancesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_RUNTIMEPROJECTATTACHMENTS = (
      'projects.locations.runtimeProjectAttachments',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'runtimeProjectAttachments/{runtimeProjectAttachmentsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_SERVERS = (
      'projects.locations.servers',
      'projects/{projectsId}/locations/{locationsId}/servers',
      {},
      ['projectsId', 'locationsId'],
      True
  )

  def __init__(self, collection_name, path, flat_paths, params,
               enable_uri_parsing):
    self.collection_name = collection_name
    self.path = path
    self.flat_paths = flat_paths
    self.params = params
    self.enable_uri_parsing = enable_uri_parsing



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The publish command for Data Product Sharing."""

from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs


@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA
)
@base.DefaultUniverseOnly
class PublishDataProduct(base.Command):
  """Publish a Knowledge Catalog Data Product or BigLake tables to external partners."""

  @staticmethod
  def Args(parser):
    def GetCatalogResourceSpec():
      return concepts.ResourceSpec(
          'biglake.dataproductsharing.v1alpha.projects.catalogs',
          resource_name='catalog',
          projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
          catalogsId=concepts.ResourceParameterAttributeConfig('catalog'),
          api_version='data_product_sharing_v1alpha',
      )
    source_group = parser.add_group('source', mutex=True, required=True)
    delta_sharing_catalog_resource = presentation_specs.ResourcePresentationSpec(
        '--connection-catalog',
        GetCatalogResourceSpec(),
        'The delta sharing catalog that contains information about where the '
        'data product is published.',
        prefixes=True,
        required=True)
    irc_catalog_resource = presentation_specs.ResourcePresentationSpec(
        '--iceberg-catalog',
        GetCatalogResourceSpec(),
        'The BigLake Iceberg REST Catalog whose tables will be published.',
        prefixes=True,
        group=source_group)
    concept_parsers.ConceptParser(
        [delta_sharing_catalog_resource, irc_catalog_resource]
    ).AddToParser(parser)
    source_group.add_argument(
        '--data-product',
        help='The Knowledge Catalog Data Product to publish.',
    )
    parser.add_argument(
        '--share',
        help='The desired name of the Share as it will be published to',
        required=True,
    )
    parser.add_argument(
        '--sap-federated-identity-provider',
        help=('The resource name of the Workload Identity Federation (WIF) '
              'provider resource representing the SAP federated identity. You '
              'must manually grant this identity the necessary IAM permissions '
              '(e.g., roles/biglake.viewer) on the underlying catalog.'),
        required=True,
    )
    parser.add_argument(
        '--title',
        help=(
            'The title of the published share. Required when publishing an'
            ' Iceberg catalog.'
        ),
    )
    parser.add_argument(
        '--short-description',
        help=(
            'The short, concise description of the published share. Required'
            ' when publishing an Iceberg catalog.'
        ),
    )
    parser.add_argument(
        '--description',
        help='The optional detailed description of the published share.',
    )

  def Run(self, args):
    client = apis.GetClientInstance('biglake', 'data_product_sharing_v1alpha')
    messages = client.MESSAGES_MODULE
    iceberg_catalog_ref = None
    if args.iceberg_catalog:
      iceberg_catalog_ref = messages.IcebergCatalogReference(
          catalog=args.CONCEPTS.iceberg_catalog.Parse().RelativeName(),
          title=args.title,
          shortDescription=args.short_description,
          description=args.description,
      )
    data_product_ref = None
    if args.data_product:
      data_product_ref = messages.DataProductReference(
          dataProduct=args.data_product
      )
    return client.dataproductsharing_v1alpha_projects_catalogs.PublishDataProduct(
        messages.BiglakeDataproductsharingV1alphaProjectsCatalogsPublishDataProductRequest(
            connectionCatalog=args.CONCEPTS.connection_catalog.Parse().RelativeName(),
            publishDataProductRequest=messages.PublishDataProductRequest(
                share=args.share,
                dataProduct=data_product_ref,
                icebergCatalog=iceberg_catalog_ref,
                sapFederatedIdentityProvider=args.sap_federated_identity_provider,
            ),
        )
    )



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from cloudsdk.google.protobuf import field_mask_pb2  # type: ignore
from cloudsdk.google.protobuf import timestamp_pb2  # type: ignore
from googlecloudsdk.generated_clients.gapic_clients.aiplatform_v1.types import operation


__protobuf__ = proto.module(
    package='google.cloud.aiplatform.v1',
    manifest={
        'SemanticGovernancePolicyEngine',
        'GetSemanticGovernancePolicyEngineRequest',
        'UpdateSemanticGovernancePolicyEngineRequest',
        'UpdateSemanticGovernancePolicyEngineOperationMetadata',
        'DeprovisionSemanticGovernancePolicyEngineRequest',
        'DeprovisionSemanticGovernancePolicyEngineOperationMetadata',
        'GatewayConfig',
    },
)


class SemanticGovernancePolicyEngine(proto.Message):
    r"""Define a singleton SemanticGovernancePolicyEngine resource
    under a project and location.

    Attributes:
        name (str):
            Identifier. The resource name of the
            SemanticGovernancePolicyEngine. Format:

            projects/{project}/locations/{location}/semanticGovernancePolicyEngine
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this
            SemanticGovernancePolicyEngine was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this
            SemanticGovernancePolicyEngine was last updated.
        psc_service_attachment (str):
            Output only. URI of the PSC attachment resource provided by
            SGP. Format:
            projects/{project}/regions/{region}/serviceAttachments/{service_attachment}
        ip_address (str):
            Output only. The private IPv4 address of the
            PSC endpoint.
        psc_forwarding_rule (str):
            Output only. The URI of the PSC endpoint resource created in
            customer project. Format:
            projects/{project}/regions/{region}/forwardingRules/{forwarding_rule}
        state (googlecloudsdk.generated_clients.gapic_clients.aiplatform_v1.types.SemanticGovernancePolicyEngine.State):
            Output only. The state of the
            SemanticGovernancePolicyEngine.
        gateway_configs (MutableMapping[str, googlecloudsdk.generated_clients.gapic_clients.aiplatform_v1.types.GatewayConfig]):
            Optional. Configurations for gateways. The
            keys are user-defined names for each gateway. At
            most 5 gateway configurations are allowed.
    """
    class State(proto.Enum):
        r"""State of the SemanticGovernancePolicyEngine.

        The lifecycle is: INACTIVE -> PROVISIONING -> {ACTIVE, FAILED}
        and ACTIVE -> DEPROVISIONING -> INACTIVE.  A FAILED engine may
        be either re-provisioned or deprovisioned.

        Values:
            STATE_UNSPECIFIED (0):
                Default value. This value is unused.
            PROVISIONING (1):
                A provisioning operation is in progress.  The
                engine will transition to ACTIVE on success or
                FAILED on failure.
            ACTIVE (2):
                The engine and all of its gateway
                configurations are provisioned and ready to
                serve traffic.
            DEPROVISIONING (3):
                A deprovisioning operation is in progress.
                The engine will transition to INACTIVE on
                success or FAILED on failure.
            INACTIVE (4):
                The engine has no provisioned infrastructure:
                either never provisioned, or successfully
                deprovisioned.
            FAILED (5):
                The most recent provisioning or
                deprovisioning operation failed. The engine may
                have partial infrastructure that needs explicit
                deprovision; the engine may be either
                re-provisioned or deprovisioned to recover.
        """
        STATE_UNSPECIFIED = 0
        PROVISIONING = 1
        ACTIVE = 2
        DEPROVISIONING = 3
        INACTIVE = 4
        FAILED = 5

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=5,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=6,
        message=timestamp_pb2.Timestamp,
    )
    psc_service_attachment: str = proto.Field(
        proto.STRING,
        number=7,
    )
    ip_address: str = proto.Field(
        proto.STRING,
        number=8,
    )
    psc_forwarding_rule: str = proto.Field(
        proto.STRING,
        number=9,
    )
    state: State = proto.Field(
        proto.ENUM,
        number=11,
        enum=State,
    )
    gateway_configs: MutableMapping[str, 'GatewayConfig'] = proto.MapField(
        proto.STRING,
        proto.MESSAGE,
        number=14,
        message='GatewayConfig',
    )


class GetSemanticGovernancePolicyEngineRequest(proto.Message):
    r"""Request message for GetSemanticGovernancePolicyEngine.

    Attributes:
        name (str):
            Required. The resource name of the
            SemanticGovernancePolicyEngine to retrieve.
            Format:

            projects/{project}/locations/{location}/semanticGovernancePolicyEngine
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class UpdateSemanticGovernancePolicyEngineRequest(proto.Message):
    r"""Request message for UpdateSemanticGovernancePolicyEngine.

    Attributes:
        semantic_governance_policy_engine (googlecloudsdk.generated_clients.gapic_clients.aiplatform_v1.types.SemanticGovernancePolicyEngine):
            Required. The SemanticGovernancePolicyEngine resource to
            update. The name field of the
            semantic_governance_policy_engine must be of the form
            ``projects/{project}/locations/{location}/semanticGovernancePolicyEngine``.
        update_mask (google.protobuf.field_mask_pb2.FieldMask):
            Optional. Specifies the fields to be overwritten in the
            SemanticGovernancePolicyEngine resource by the update. The
            fields specified in the update_mask are relative to the
            resource itself. If no update_mask is provided, all fields
            are overwritten.
    """

    semantic_governance_policy_engine: 'SemanticGovernancePolicyEngine' = proto.Field(
        proto.MESSAGE,
        number=1,
        message='SemanticGovernancePolicyEngine',
    )
    update_mask: field_mask_pb2.FieldMask = proto.Field(
        proto.MESSAGE,
        number=2,
        message=field_mask_pb2.FieldMask,
    )


class UpdateSemanticGovernancePolicyEngineOperationMetadata(proto.Message):
    r"""Details of
    [SemanticGovernancePolicyEngineService.UpdateSemanticGovernancePolicyEngine][google.cloud.aiplatform.v1.SemanticGovernancePolicyEngineService.UpdateSemanticGovernancePolicyEngine]
    operation.

    Attributes:
        generic_metadata (googlecloudsdk.generated_clients.gapic_clients.aiplatform_v1.types.GenericOperationMetadata):
            The common part of the operation metadata.
        progress_message (str):
            Output only. Granular progress message for
            the operation.
    """

    generic_metadata: operation.GenericOperationMetadata = proto.Field(
        proto.MESSAGE,
        number=1,
        message=operation.GenericOperationMetadata,
    )
    progress_message: str = proto.Field(
        proto.STRING,
        number=2,
    )


class DeprovisionSemanticGovernancePolicyEngineRequest(proto.Message):
    r"""Request message for
    [SemanticGovernancePolicyEngineService.DeprovisionSemanticGovernancePolicyEngine][google.cloud.aiplatform.v1.SemanticGovernancePolicyEngineService.DeprovisionSemanticGovernancePolicyEngine].

    Attributes:
        name (str):
            Required. The resource name of the
            SemanticGovernancePolicyEngine to deprovision.
            Format:

            projects/{project}/locations/{location}/semanticGovernancePolicyEngine
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class DeprovisionSemanticGovernancePolicyEngineOperationMetadata(proto.Message):
    r"""Details of
    [SemanticGovernancePolicyEngineService.DeprovisionSemanticGovernancePolicyEngine][google.cloud.aiplatform.v1.SemanticGovernancePolicyEngineService.DeprovisionSemanticGovernancePolicyEngine]
    operation.

    Attributes:
        generic_metadata (googlecloudsdk.generated_clients.gapic_clients.aiplatform_v1.types.GenericOperationMetadata):
            The common part of the operation metadata.
    """

    generic_metadata: operation.GenericOperationMetadata = proto.Field(
        proto.MESSAGE,
        number=1,
        message=operation.GenericOperationMetadata,
    )


class GatewayConfig(proto.Message):
    r"""Configuration for a single gateway.

    Attributes:
        network (str):
            Optional. The URI of the network resource where PSC-E will
            be provisioned. if not provided ``default`` network will be
            used. Format: projects/{project}/global/networks/{network}
        subnetwork (str):
            Optional. The URI of the subnetwork resource where PSC-E
            will be provisioned. if not provided ``default`` subnet will
            be used from the same {location} Format:
            projects/{project}/regions/{region}/subnetworks/{subnetwork}
        dns_zone_name (str):
            Optional. FQDN of the private DNS zone to
            create DNS record set for PSC endpoint.
        state (googlecloudsdk.generated_clients.gapic_clients.aiplatform_v1.types.GatewayConfig.State):
            Output only. The state of the Gateway
            configuration.
        ip_address (str):
            Output only. The private IP address of the
            PSC endpoint.
        psc_endpoint (str):
            Output only. The self-link or name of the
            Private Service Connect endpoint forwarding
            rule.
        dns_record (str):
            Output only. The fully qualified record name
            of the created A-record in Cloud DNS.
    """
    class State(proto.Enum):
        r"""State of the Gateway configuration.

        Values:
            STATE_UNSPECIFIED (0):
                The default value. This value is used if the
                state is omitted.
            PROVISIONING (1):
                The Gateway is being provisioned.
            ACTIVE (2):
                The Gateway is active and ready to use.
            DEPROVISIONING (3):
                The Gateway is being de-provisioned.
            INACTIVE (4):
                The Gateway is inactive.
            FAILED (5):
                The Gateway failed to be provisioned.
        """
        STATE_UNSPECIFIED = 0
        PROVISIONING = 1
        ACTIVE = 2
        DEPROVISIONING = 3
        INACTIVE = 4
        FAILED = 5

    network: str = proto.Field(
        proto.STRING,
        number=2,
    )
    subnetwork: str = proto.Field(
        proto.STRING,
        number=3,
    )
    dns_zone_name: str = proto.Field(
        proto.STRING,
        number=4,
    )
    state: State = proto.Field(
        proto.ENUM,
        number=5,
        enum=State,
    )
    ip_address: str = proto.Field(
        proto.STRING,
        number=6,
    )
    psc_endpoint: str = proto.Field(
        proto.STRING,
        number=7,
    )
    dns_record: str = proto.Field(
        proto.STRING,
        number=8,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library for obtaining API clients and messages.

This should only be called by api_lib.util.apis, core.resources, gcloud meta
commands, and module tests.
"""


from googlecloudsdk.api_lib.util import apis_util
from googlecloudsdk.api_lib.util import resource as resource_util
from googlecloudsdk.core import exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core import requests as core_requests
from googlecloudsdk.core import transport
from googlecloudsdk.core.util import regional
from googlecloudsdk.generated_clients.apis import apis_map
import requests
import six
from six.moves.urllib.parse import urljoin
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import urlunparse
import urllib3


def _GetApiNameAndAlias(api_name):
  # pylint:disable=protected-access
  return (apis_util._API_NAME_ALIASES.get(api_name, api_name), api_name)


def _GetDefaultVersion(api_name):
  api_name, _ = _GetApiNameAndAlias(api_name)
  api_vers = apis_map.MAP.get(api_name, {})
  for ver, api_def in six.iteritems(api_vers):
    if api_def.default_version:
      return ver
  return None


def _GetApiNames():
  """Returns list of avaiblable apis, ignoring the version."""
  return sorted(apis_map.MAP.keys())


def _GetVersions(api_name):
  """Return available versions for given api.

  Args:
    api_name: str, The API name (or the command surface name, if different).

  Raises:
    apis_util.UnknownAPIError: If api_name does not exist in the APIs map.

  Returns:
    list, of version names.
  """
  api_name, _ = _GetApiNameAndAlias(api_name)
  version_map = apis_map.MAP.get(api_name, None)
  if version_map is None:
    raise apis_util.UnknownAPIError(api_name)
  return list(version_map.keys())


def GetApiDef(api_name, api_version):
  """Returns the APIDef for the specified API and version.

  Args:
    api_name: str, The API name (or the command surface name, if different).
    api_version: str, The version of the API.

  Raises:
    apis_util.UnknownAPIError: If api_name does not exist in the APIs map.
    apis_util.UnknownVersionError: If api_version does not exist for given
      api_name in the APIs map.

  Returns:
    APIDef, The APIDef for the specified API and version.
  """
  api_name, api_name_alias = _GetApiNameAndAlias(api_name)
  if api_name not in apis_map.MAP:
    raise apis_util.UnknownAPIError(api_name)

  version_overrides = properties.VALUES.api_client_overrides.AllValues()

  # First attempt to get api specific override, then full surface override.
  version_override = version_overrides.get('{}/{}'.format(
      api_name, api_version))
  if not version_override:
    version_override = version_overrides.get(api_name_alias, None)

  api_version = version_override or api_version

  api_versions = apis_map.MAP[api_name]
  if api_version is None or api_version not in api_versions:
    raise apis_util.UnknownVersionError(api_name, api_version)
  else:
    api_def = api_versions[api_version]

  return api_def


def _GetClientClass(api_name, api_version):
  """Returns the client class for the API specified in the args.

  Args:
    api_name: str, The API name (or the command surface name, if different).
    api_version: str, The version of the API.

  Returns:
    base_api.BaseApiClient, Client class for the specified API.
  """
  api_def = GetApiDef(api_name, api_version)
  return _GetClientClassFromDef(api_def)


def _GetClientClassFromDef(api_def):
  """Returns the apitools client class for the API definition specified in args.

  Args:
    api_def: apis_map.APIDef, The definition of the API.

  Returns:
    base_api.BaseApiClient, Client class for the specified API.
  """
  client_full_classpath = api_def.apitools.client_full_classpath
  module_path, client_class_name = client_full_classpath.rsplit('.', 1)
  module_obj = __import__(module_path, fromlist=[client_class_name])
  return getattr(module_obj, client_class_name)


def _GetClientInstance(api_name,
                       api_version,
                       no_http=False,
                       http_client=None,
                       check_response_func=None,
                       http_timeout_sec=None,
                       region=None):
  """Returns an instance of the API client specified in the args.

  Args:
    api_name: str, The API name (or the command surface name, if different).
    api_version: str, The version of the API.
    no_http: bool, True to not create an http object for this client.
    http_client: bring your own http client to use. Incompatible with
      no_http=True.
    check_response_func: error handling callback to give to apitools.
    http_timeout_sec: int, seconds of http timeout to set, defaults if None.
    region: str, Region (or multi-region) for regionalized endpoints (REP).

  Returns:
    base_api.BaseApiClient, An instance of the specified API client.
  """

  # pylint: disable=g-import-not-at-top
  if no_http:
    assert http_client is None
  elif http_client is None:
    # Normal gcloud authentication
    # Import http only when needed, as it depends on credential infrastructure
    # which is not needed in all cases.
    from googlecloudsdk.core.credentials import transports
    http_client = transports.GetApitoolsTransport(
        response_encoding=transport.ENCODING,
        timeout=http_timeout_sec if http_timeout_sec else 'unset')
    # Add exception handling for potential DNS resolution failures due to
    # unknown regional endpoints.
    if region and _ShouldUseRegionalEndpoints(api_name, api_version, region):
      api_def = GetApiDef(api_name, api_version)
      if region not in api_def.regional_endpoints:
        def _UnavailableRegionDnsExceptionHandler(e):
          # requests.exceptions.ConnectionError wraps
          # urllib3.exceptions.MaxRetryError which wraps
          # urllib3.exceptions.NameResolutionError.
          if (
              e.args
              and isinstance(e.args[0], urllib3.exceptions.MaxRetryError)
              and isinstance(
                  e.args[0].reason, urllib3.exceptions.NameResolutionError)):
            raise regional.UnavailableRegionError(
                '{}\n\nNote: the region [{}] may not be available for this '
                'service. Known available regions are: [{}].'.format(
                    e, region, ', '.join(sorted(api_def.regional_endpoints))))
          raise e
        core_requests.RequestWrapper().WrapRequest(
            http_client.session,
            [],
            exc_handler=_UnavailableRegionDnsExceptionHandler,
            exc_type=requests.exceptions.ConnectionError,
        )

  client_class = _GetClientClass(api_name, api_version)
  if no_http:
    endpoint = client_class.BASE_URL
  else:
    endpoint = _GetEffectiveApiEndpoint(
        api_name, api_version, client_class, region)

  client_instance = client_class(
      url=endpoint,
      get_credentials=False,
      http=http_client)
  if check_response_func is not None:
    client_instance.check_response_func = check_response_func
  api_key = properties.VALUES.core.api_key.Get()
  if api_key:
    client_instance.AddGlobalParam('key', api_key)
    header = 'X-Google-Project-Override'
    client_instance.additional_http_headers[header] = 'apikey'
  return client_instance


def _GetGapicClientClass(api_name,
                         api_version,
                         transport_choice=apis_util.GapicTransport.GRPC):
  """Returns the GAPIC client class for the API def specified by the args.

  Args:
    api_name: str, The API name (or the command surface name, if different).
    api_version: str, The version of the API.
    transport_choice: apis_util.GapicTransport, The transport to be used by the
      client.
  """
  api_def = GetApiDef(api_name, api_version)
  if transport_choice == apis_util.GapicTransport.GRPC_ASYNCIO:
    client_full_classpath = api_def.gapic.async_client_full_classpath
  elif transport_choice == apis_util.GapicTransport.REST:
    client_full_classpath = api_def.gapic.rest_client_full_classpath
  else:
    client_full_classpath = api_def.gapic.client_full_classpath

  module_path, client_class_name = client_full_classpath.rsplit('.', 1)
  module_obj = __import__(module_path, fromlist=[client_class_name])
  return getattr(module_obj, client_class_name)


def _GetGapicClientInstance(
    api_name,
    api_version,
    credentials,
    address_override_func=None,
    transport_choice=apis_util.GapicTransport.GRPC,
    attempt_direct_path=False,
    redact_request_body_reason=None,
    custom_interceptors=None,
    region=None,
    channel_options=None,
):
  """Returns an instance of the GAPIC API client specified in the args.

  For apitools API clients, the API endpoint override is something like
  http://compute.googleapis.com/v1/. For GAPIC API clients, the DEFAULT_ENDPOINT
  is something like compute.googleapis.com. To use the same endpoint override
  property for both, we use the netloc of the API endpoint override.

  Args:
    api_name: str, The API name (or the command surface name, if different).
    api_version: str, The version of the API.
    credentials: google.auth.credentials.Credentials, the credentials to use.
    address_override_func: function, function to call to override the client
      host. It takes a single argument which is the original host.
    transport_choice: apis_util.GapicTransport, The transport to be used by the
      client.
    attempt_direct_path: bool, True if we want to attempt direct path gRPC where
      possible.
    redact_request_body_reason: str, the reason why the request body must be
      redacted if --log-http is used. If None, the body is not redacted.
    custom_interceptors: list[grpc interceptor], a list of custom interceptors
    to add to the channel.
    region: str, Region (or multi-region) for regionalized endpoints (REP).
    channel_options: dict, optional channel arguments to pass to the underlying
      gRPC channel.

  Returns:
    An instance of the specified GAPIC API client.
  """
  def AddressOverride(address):
    try:
      endpoint_override = properties.VALUES.api_endpoint_overrides.Property(
          api_name).Get()
    except properties.NoSuchPropertyError:
      endpoint_override = None

    if endpoint_override:
      address = urlparse(endpoint_override).netloc

    if address_override_func:
      address = address_override_func(address)

    if endpoint_override is not None:
      return address

    # Note that regionalized endpoints are incompatible with non-Google
    # universe domains.
    if _ShouldUseRegionalEndpoints(api_name, api_version, region):
      return _GetRegionalizedEndpoint(address, region)

    return UniversifyAddress(address)

  # Add exception handling for potential DNS resolution failures due to unknown
  # regional endpoints.
  region_interceptors = []
  if region and _ShouldUseRegionalEndpoints(api_name, api_version, region):
    # pylint: disable=g-import-not-at-top
    from googlecloudsdk.core import gapic_util_internal
    # pylint: enable=g-import-not-at-top
    api_def = GetApiDef(api_name, api_version)
    if region not in api_def.regional_endpoints:
      interceptor = gapic_util_internal.UnavailableRegionDnsErrorInterceptor(
          region, known_available_regions=api_def.regional_endpoints)
      region_interceptors.append(interceptor)

  client_class = _GetGapicClientClass(
      api_name, api_version, transport_choice=transport_choice)

  return client_class(
      credentials,
      address_override_func=AddressOverride,
      mtls_enabled=_MtlsEnabled(api_name, api_version),
      attempt_direct_path=attempt_direct_path,
      redact_request_body_reason=redact_request_body_reason,
      custom_interceptors=(custom_interceptors or []) + region_interceptors,
      channel_options=channel_options,
  )


def UniversifyAddress(address):
  """Update a URL based on the current universe domain."""
  universe_domain_property = properties.VALUES.core.universe_domain
  universe_domain = universe_domain_property.Get()
  if (address is not None and
      universe_domain_property.default != universe_domain):
    address = address.replace(universe_domain_property.default,
                              universe_domain, 1)
  return address


def _GetRegionalizedEndpoint(domain, region):
  """Returns regionalized domain.

  Regional endpoints (REPs) have the following format:
    {service}.{region}.rep.googleapis.com

  Args:
    domain: Domain for the API in form {service}.googleapis.com.
    region: REP region (or multi-region).

  Returns:
    str, Regionalized domain.
  """
  if region is None:
    raise ValueError('Region must be provided.')
  service, base_domain = domain.split('.', 1)
  match base_domain:
    case 'googleapis.com' | 'mtls.googleapis.com':
      return f'{service}.{region}.rep.{base_domain}'
    case _:
      raise ValueError(
          'Base URLs should have form {service}.googleapis.com;'
          f' received: [{domain}].'
      )


class UnsupportedEndpointModeError(exceptions.Error):
  """Error when regional/endpoint_mode property is unsupported by a command."""


def _ShouldUseRegionalEndpoints(api_name, api_version, region):
  """Returns whether or not regional endpoint should be used.

  Depending on the user-specified regional/endpoint_mode setting and the
  developer-specified command annotation, this function calculates whether or
  not to use regional endpoints. In case of an invalid combination, an error
  will be raised. The full matrix is as follows:

  +--------------+-------------------------------------------------------------+
  |              |                    Command compatibility                    |
  |  User mode   +---------------+---------------+-----------------------------+
  |              | Only supports | Only supports | Supports both               |
  |              | global        | regional      |                             |
  +--------------+---------------+---------------+-----------------------------+
  | regional     | Error         | Regional      | Regional client             |
  |              |               | client        |                             |
  +--------------+---------------+---------------+-----------------------------+
  | global       | Global client | Error         | Global client               |
  +--------------+---------------+---------------+-----------------------------+
  | regional-    | Global client | Regional      | If region available:        |
  | preferred    |               | client        | Regional client             |
  |              |               |               | Otherwise: Global client    |
  +--------------+---------------+---------------+-----------------------------+
  | (unset)      | Global client | Regional      | Global client               |
  |              |               | client        |                             |
  +--------------+---------------+---------------+-----------------------------+

  Args:
    api_name: str, The API name.
    api_version: str, The version of the API.
    region: The region (or multi-region).
  Returns:
    bool, Whether to use regional endpoints.
  Raises:
    UnsupportedEndpointModeError: If the command does not support the given
      endpoint mode.
  """
  mode = properties.VALUES.regional.endpoint_mode.Get()
  compatibility = properties.VALUES.regional.endpoint_compatibility.Get()
  if mode == properties.VALUES.regional.REGIONAL and compatibility is None:
    raise UnsupportedEndpointModeError(
        f'The regional/endpoint_mode property is currently set to [{mode}], but'
        ' this command does not support regional endpoints.')
  if (
      mode == properties.VALUES.regional.GLOBAL and
      compatibility == properties.VALUES.regional.REQUIRED
  ):
    raise UnsupportedEndpointModeError(
        f'The regional/endpoint_mode property is currently set to [{mode}], but'
        ' this command only supports regional endpoints.')

  # Global.
  if (
      compatibility is None
      or mode == properties.VALUES.regional.GLOBAL
      or (
          mode is None
          and compatibility == properties.VALUES.regional.SUPPORTED
      )
  ):
    return False

  # Regional if available.
  known_available_regions = GetApiDef(api_name, api_version).regional_endpoints
  if (
      mode == properties.VALUES.regional.REGIONAL_PREFERRED
      and compatibility == properties.VALUES.regional.SUPPORTED
  ):
    return region in known_available_regions

  # Command requires or user explicitly requested regional endpoints.
  if region not in known_available_regions:
    # If the requested region isn't known to be available, it could be because
    # this version of gcloud is older and doesn't know about it yet. We let the
    # request through and fail open in this case rather than error out here.
    log.info(
        'API [%s] does not contain a known regional endpoint for region [%s].',
        api_name, region)
  return True


def _GetMtlsEndpoint(api_name, api_version, client_class=None):
  """Returns mtls endpoint."""
  api_def = GetApiDef(api_name, api_version)
  if api_def.apitools:
    client_class = client_class or _GetClientClass(api_name, api_version)
  else:
    client_class = client_class or _GetGapicClientClass(api_name, api_version)
  return api_def.mtls_endpoint_override or client_class.MTLS_BASE_URL


def _MtlsEnabled(api_name, api_version):
  """Checks if the API of the given version should use mTLS.

  If context_aware/always_use_mtls_endpoint is True, then mTLS will always be
  used.

  If context_aware/use_client_certificate is True, then mTLS will be used only
  if the API version is in the mTLS allowlist.

  gcloud maintains a client-side allowlist for the mTLS feature
  (go/gcloud-rollout-mtls).

  Args:
    api_name: str, The API name.
    api_version: str, The version of the API.

  Returns:
    True if the given service and version is in the mTLS allowlist.
  """
  if properties.VALUES.context_aware.always_use_mtls_endpoint.GetBool():
    return True

  if not properties.VALUES.context_aware.use_client_certificate.GetBool():
    return False

  api_def = GetApiDef(api_name, api_version)
  return api_def.enable_mtls


def _BuildEndpointOverride(endpoint_override, base_url):
  """Constructs a normalized endpoint URI depending on the client base_url."""
  url_base = urlparse(base_url)
  url_endpoint_override = urlparse(endpoint_override)
  if url_base.path == '/' or url_endpoint_override.path != '/':
    return endpoint_override
  return urljoin(
      '{}://{}'.format(url_endpoint_override.scheme,
                       url_endpoint_override.netloc), url_base.path)


def _GetBaseUrlFromApi(api_name, api_version):
  """Returns base url for given api."""
  if GetApiDef(api_name, api_version).apitools:
    client_class = _GetClientClass(api_name, api_version)
  else:
    client_class = _GetGapicClientClass(api_name, api_version)

  if hasattr(client_class, 'BASE_URL'):
    client_base_url = client_class.BASE_URL
  else:
    try:
      client_base_url = _GetResourceModule(api_name, api_version).BASE_URL
    except AttributeError:
      client_base_url = 'https://{}.googleapis.com/{}'.format(
          api_name, api_version
      )
  return UniversifyAddress(client_base_url)


def _GetEffectiveApiEndpoint(
    api_name, api_version, client_class=None, region=None):
  """Returns effective endpoint for given api."""
  try:
    endpoint_override = properties.VALUES.api_endpoint_overrides.Property(
        api_name).Get()
  except properties.NoSuchPropertyError:
    endpoint_override = None

  api_def = GetApiDef(api_name, api_version)
  if api_def.apitools:
    client_class = client_class or _GetClientClass(api_name, api_version)
  else:
    client_class = client_class or _GetGapicClientClass(api_name, api_version)
  client_base_url = _GetBaseUrlFromApi(api_name, api_version)
  if endpoint_override:
    address = _BuildEndpointOverride(endpoint_override, client_base_url)
  # Note that regionalized endpoints are incompatible with non-Google
  # universe domains.
  elif _ShouldUseRegionalEndpoints(api_name, api_version, region):
    if _MtlsEnabled(api_name, api_version):
      client_base_url = _GetMtlsEndpoint(api_name, api_version, client_class)
    url_base = urlparse(client_base_url)
    regionalized_domain = _GetRegionalizedEndpoint(url_base.netloc, region)
    address = urlunparse(url_base._replace(netloc=regionalized_domain))
  elif _MtlsEnabled(api_name, api_version):
    address = UniversifyAddress(
        _GetMtlsEndpoint(api_name, api_version, client_class)
    )
  else:
    address = client_base_url

  return address


def _GetMessagesModule(api_name, api_version):
  """Returns the messages module for the API specified in the args.

  Args:
    api_name: str, The API name (or the command surface name, if different).
    api_version: str, The version of the API.

  Returns:
    Module containing the definitions of messages for the specified API.
  """
  api_def = GetApiDef(api_name, api_version)
  # fromlist below must not be empty, see:
  # http://stackoverflow.com/questions/2724260/why-does-pythons-import-require-fromlist.
  return __import__(
      api_def.apitools.messages_full_modulepath, fromlist=['something'])


def _GetResourceModule(api_name, api_version):
  """Imports and returns given api resources module."""

  api_def = GetApiDef(api_name, api_version)
  # fromlist below must not be empty, see:
  # http://stackoverflow.com/questions/2724260/why-does-pythons-import-require-fromlist.
  if api_def.apitools:
    return __import__(
        api_def.apitools.class_path + '.' + 'resources', fromlist=['something']
    )
  # ApiDef must be gapic-only:
  return __import__(
      api_def.gapic.class_path + '.' + 'resources', fromlist=['something']
  )


def _GetApiCollections(api_name, api_version):
  """Yields all collections for for given api."""

  try:
    resources_module = _GetResourceModule(api_name, api_version)
  except ImportError:
    pass
  else:
    for collection in resources_module.Collections:
      yield resource_util.CollectionInfo(
          api_name,
          api_version,
          resources_module.BASE_URL,
          resources_module.DOCS_URL,
          collection.collection_name,
          collection.path,
          collection.flat_paths,
          collection.params,
          collection.enable_uri_parsing,
      )
